#!/usr/bin/env python

"""Data file structure and i/o.

This is the (qubx.tree) format of a simulation scratch file::

 Sampling
 SignalCount
 Signals
   [] Signal
     Name
     Units
 Segments
   [] Segment
     SampleCount
     Signals
       [] Signal
         Samples [(float[SampleCount])]
 IdealSignals
   [] IdealSignal
     SignalIndex
     Segments
       [] Segment [f, l] {amp, sd, DwellCount, Classes, Firsts, Lasts}


Copyright 2008-2014 Research Foundation State University of New York 
This file is part of QUB Express.                                          

QUB Express is free software; you can redistribute it and/or modify          
it under the terms of the GNU General Public License as published by 
the Free Software Foundation, either version 3 of the License, or    
(at your option) any later version.                                  

QUB Express is distributed in the hope that it will be useful,               
but WITHOUT ANY WARRANTY; without even the implied warranty of       
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
GNU General Public License for more details.                         

You should have received a copy of the GNU General Public License,   
named LICENSE.txt, in the QUB Express program directory.  If not, see        
<http://www.gnu.org/licenses/>.                                      

"""

import bisect
import collections
import itertools
import mmap
import os
import numpy
import re
import tempfile
from itertools import izip, count

import qubx.fast.filter
from qubx.fast.baseline import BaselineSeg
import qubx.tree
import qubx.trialset
from qubx.ideal import Idealization, IdlFits
from qubx.util_types import *
from qubx.table import *
from qubx.table_tree import *
from qubx.undo import UndoStack

from .sorted_collection import SortedCollection

from ctypes import c_int, c_ushort, c_short, c_double, c_float, POINTER, byref
c_int_p = POINTER(c_int)
c_float_p = POINTER(c_float)

DATASOURCE_FILE, DATASOURCE_SCREEN, DATASOURCE_LIST = (0,1,2)

DOWNSAMPLE_ALL_LOGCOUNT = 2.0
DOWNSAMPLE_LOGSLOPE = 3.0
DOWNSAMPLES_PER_PIX = lambda Nsigseg, Nsamp: exp( (log(Nsigseg*Nsamp) - DOWNSAMPLE_ALL_LOGCOUNT) / DOWNSAMPLE_LOGSLOPE + DOWNSAMPLE_ALL_LOGCOUNT )

SAVEBUFLEN = 2**18

Readers = {}
def SetReader(ext, description, create):
    rtn = None
    if ext.lower() in Readers:
        rtn = Readers[ext.lower()].create
    Readers[ext.lower()] = Anon(ext=ext, pattern='*%s'%ext, description=description, create=create)
    return rtn

def GetReader(ext):
    try:
        return Readers[ext.lower()]
    except KeyError:
        return Readers['.txt']

def Open(path, progressf=lambda frac: True):
    try:
        data = GetReader(os.path.splitext(path)[1]).create(path, progressf)
    except:
        traceback.print_exc()
        data = GetReader('.txt').create(path, progressf)
    setup_segments_list(data)
    return data

def setup_segments_list(data):
    lseg = data.lists.by_name('Segments')
    if lseg.size == len(data.segmentation.segments):
        i = 0
        for f, l, n in data.segmentation.segments:
            if (lseg[i, 'From'] != f) or (lseg[i, 'To'] != l):
                break
            i += 1
        else: # for loop not broken out of: all match
            return
    lseg.clear()
    lseg.insert_selections([(fln[0], fln[1],
                             {'Duration':fln[2]*1e3*data.sampling, 'SampleCount':fln[2], 'Start':data.segmentation.starts[i]})
                            for i, fln in enumerate(data.segmentation.segments)])
    lseg.user_can_edit = False


Writers = {}
def SetWriter(ext, description, save):
    Writers[ext.lower()] = Anon(ext=ext, pattern='*%s'%ext, description=description, save=save)

def Save(path, *args, **kw):
    ext = os.path.splitext(path)[1].lower()
    try:
        save = Writers[ext].save
    except KeyError:
        raise IOError('Unknown data format: %s'%ext)
    return save(path, *args, **kw)


# replace with gui functions in graphical unit:
RequestSampling = lambda s: s
RequestBinaryParams = lambda sampling, scaling, floating, bytes, signals: (sampling, scaling, floating, bytes, signals)




class Segmentation(object):
    """Describes how a contiguous range of samples is divided into segments.

    @ivar segments: list of (first, last, sample_count) of each segment
    @ivar starts: list of start [ms] of each segment
    @ivar idl: L{qubx.ideal.Idealization} with one dwell per segment, for quick index_at() 
    @ivar OnAdd: L{WeakEvent}(Segmentation, first, last, start) when add_seg() is called
    @ivar OnClear: L{WeakEvent}(Segmentation) when clear() is called
"""
    def __init__(self):
        self.OnAdd = WeakEvent() # (Segmentation, f, l)
        self.OnClear = WeakEvent() # (Segmentation)
        self.clear()
    def clear(self):
        """Removes all segments."""
        self.segments = []
        self.starts = []
        self.last = -1
        self.idl = Idealization(1.0)
        self.OnClear(self)
    def add_seg(self, f, l, start):
        """Appends a segment."""
        if l > self.last:
            self.idl.add_seg(self.last+1, l)
            self.last = l
        seg_ix = len(self.segments)
        self.segments.append( (f, l, l-f+1) )
        self.starts.append(start)
        self.idl.set_dwell(f, l, seg_ix)
        self.OnAdd(self, f, l, start)
    def index_at(self, i):
        """Returns the index of the segment containing sample i."""
        return self.idl[i]
    def split(self, f, l):
        """Returns segmented list of (iseg, f, l) with local segment indices."""
        chunks = []
        while f <= l:
            iseg = self.index_at(f)
            segf, segl, segn = self.segments[iseg]
            take = min(l, segl) - f + 1
            if take <= 0:
                if qubx.global_namespace.DEBUG:
                    print 'segmentation split ended early',f, l, segl, iseg
                break
            chunks.append((iseg, f-segf, f-segf+take-1))
            f += take
        return chunks

class SigBaseline(object):
    """Represents piecewise-linear baseline of one signal."""
    def __init__(self, segmentation):
        self.__ref = Reffer()
        self.segs = []
        self.OnChange = WeakEvent()
        self.segm = segmentation
        self.segm.OnAdd += self.__ref(self.__onAddSeg)
        self.segm.OnClear += self.__ref(self.__onClearSegs)
        self.__onClearSegs(self.segm)
        for f, l, n in self.segm.segments:
            self.__onAddSeg(segmentation, f, l)
    def __onAddSeg(self, segm, f, l, start=0.0):
        self.segs.append(BaselineSeg(f, l))
    def __onClearSegs(self, segm):
        self.segs = []
    def find_seg(self, f):
        return self.segs[self.segm.index_at(f)]
    def first_node(self, seg_ix):
        return self.segs[seg_ix].first_node()
    def last_node(self, seg_ix):
        return self.segs[seg_ix].last_node()
    def interpolate(self, point):
        return self.find_seg(point).interpolate(point)
    def range_at(self, point):
        return self.find_seg(point).range_at(point)
    def add_node(self, point, value):
        self.find_seg(point).add_node(point, value)
        self.OnChange()
    def add_nodes(self, points, values):
        if not len(points): return
        p = 0
        for iseg, f, l in self.segm.split(points[0], points[-1]):
            segf = self.segm.segments[iseg][0]
            q = p
            while (q < len(points)) and (points[p] <= (segf+l)):
                q += 1
            self.segs[iseg].add_nodes(points[p:q], values[p:q])
        self.OnChange()
    def clear_nodes(self, first, last):
        p = 0
        for iseg, f, l in self.segm.split(first, last):
            segf = self.segm.segments[iseg][0]
            self.segs[iseg].clear_nodes(segf+f, segf+l)
        self.OnChange()
    def clear_all(self):
        for i, fln in enumerate(self.segm.segments):
            f, l, n = fln
            self.segs[i].clear_nodes(f, l)
        self.OnChange()
    def get_sampled_nodes(self, first, last):
        samples = numpy.zeros(shape=(last-first+1,), dtype='float32')
        at = 0
        for iseg, f, l in self.segm.split(first, last):
            segf = self.segm.segments[iseg][0]
            brange = self.segs[iseg].range_at(segf+f)
            while brange and (brange.node_a.point <= (segf+l)):
                x0 = max(first, brange.node_a.point)
                x1 = min(last, brange.node_b.point)
                y0 = brange.node_a.value
                y1 = brange.node_b.value
                n = x1 - x0
                slope = (y1 - y0) / (brange.node_b.point - brange.node_a.point)
                y0 += slope * (x0 - brange.node_a.point)
                if x1 == segf + l:
                    n += 1 # to be inclusive only on final node
                samples_range = numpy.arange(n, dtype='float32')
                samples_range *= slope
                samples_range += y0
                samples[at:at+n] = samples_range
                at += n
                brange = brange.next()
        return samples

            
class SigIdl(object):
    """Represents one idealized signal; resizes with L{Segmentation}; stores per-segment info.

@ivar sampling: sampling interval, in seconds; please keep synchronized with data.sampling
@ivar idl: L{qubx.ideal.Idealization}
@ivar seg: list of L{Anon}(f=first, l=last, ix=index, amp=[amp_of_class], std=[std_of_class])
"""
    def __init__(self, sampling, segmentation, construct_idl=Idealization):
        self.ref = Reffer()
        self._sampling = sampling
        self.Idealization = construct_idl
        self.segm = segmentation
        self.segm.OnAdd += self.ref(self._onAddSeg)
        self.segm.OnClear += self.ref(self._onClearSegs)
        self._onClearSegs(self.segm)
        for f, l, n in self.segm.segments:
            self._onAddSeg(segmentation, f, l, f*sampling*1e3)
    def set_sampling(self, x):
        self._sampling = x
        self.idl.set_sampling(x*1e3)
    sampling = property(lambda self: self._sampling, lambda self, x: self.set_sampling(x))
    def _onAddSeg(self, segm, f, l, start):
        if f > (self.last+1):
            self.idl.add_seg(self.last+1, f-1)
            self.segsAndSpacers += 1
        self.last = l
        self.idl.add_seg(f, l)
        self.segix.append(self.segsAndSpacers)
        self.segsAndSpacers += 1
        self.seg.append(Anon(f=f, l=l, ix=len(self.seg), amp=[], std=[]))
    def _onClearSegs(self, segm):
        self.seg = [] # per-seg attributes
        self.idl = self.Idealization(self._sampling*1e3)
        self.segix = []
        self.segsAndSpacers = 0
        self.last = -1

class SigOverlays(SigIdl):
    def __init__(self, sampling, segmentation):
        SigIdl.__init__(self, sampling, segmentation)
        self.OnChangeSource = WeakEvent() # (self)
        self.OnChangeOverlays = WeakEvent() # (self) # QubData subscribe on creation, dispatch to its OnChangeOverlays, OnChangeOverlaySource
        self.sels = []
        self.data = []
        self.f = self.l = None
        self.cls = -1
        self.undo = qubx.undo.UndoStack()
    def set_source(self, f, l):
        self.f, self.l = f, l
        self.sels.append((f, l))
        self.data.append(None) # filled by QubData_Analog as cache
        self.cls += 1
        self.undo.push_undo(self.pop_source, bind(self.set_source, f, l))
        self.undo.seal_undo('Select Source')
        self.OnChangeSource()
    def pop_source(self):
        if self.sels:
            del self.sels[-1]
            del self.data[-1]
            if self.sels:
                self.f, self.l = self.sels[-1]
            else:
                self.f = self.l = None
            self.cls -= 1
            self.OnChangeSource()
    def pop_dest(self, n, ff, ll, cc):
        self.idl.set_dwells(n, ff, ll, cc)
        self.OnChangeOverlays()
    def add_dest(self, f, l):
        ff, ll, cc = self.idl.get_dwells_and_gaps(f, l)
        self.idl.set_dwell(f, l, self.cls)
        self.undo.push_undo(bind(self.pop_dest, len(ff), ff, ll, cc), bind(self.add_dest, f, l))
        self.undo.seal_undo('Select Destination')
        self.OnChangeOverlays()
    def restore(self, f, l):
        ff, ll, cc = self.idl.get_dwells_and_gaps(f, l)
        self.idl.set_dwell(f, l, -1)
        self.undo.push_undo(bind(self.pop_dest, len(ff), ff, ll, cc), bind(self.restore, f, l))
        self.undo.seal_undo('Restore')
        self.OnChangeOverlays()
    def clear(self):
        for i, fln in enumerate(self.segm.segments):
            f, l, n = fln
            self.idl.set_dwell(f, l, -1)
        self.data[:] = []
        self.sels[:] = []
        self.f = self.l = None
        self.cls = -1
        self.undo.clear()
        self.OnChangeSource()
        self.OnChangeOverlays()


class Exclusion(SigIdl):
    """Represents excluded and included regions as an idealization with class 0 and 1 resp.
    
    @ivar OnChange: L{WeakEvent}(Exclusion)
"""
    def __init__(self, sampling, segmentation):
        SigIdl.__init__(self, sampling, segmentation)
        self.OnChange = WeakEvent() # (Exclusion)
    def _onAddSeg(self, segm, f, l, start):
        SigIdl._onAddSeg(self, segm, f, l, start)
        self.include(f, l)
    def include(self, f, l):
        self.idl.set_dwell(f, l, 1)
        self.OnChange(self)
    def exclude(self, f, l):
        self.idl.set_dwell(f, l, 0)
        self.OnChange(self)
    def include_all(self):
        for f, l, n in self.segm.segments:
            self.idl.set_dwell(f, l, 1)
        self.OnChange(self)
    def exclude_all(self):
        for f, l, n in self.segm.segments:
            self.idl.set_dwell(f, l, 0)
        self.OnChange(self)


class SelectionList(qubx.table.SimpleTable):
    def __init__(self):
        qubx.table.SimpleTable.__init__(self, 'List', auto_add_fields=True, global_name='QubX.Data.file.list', sortable=True)
        self.__ref = Reffer()
        self.OnRemoved += self.__ref(self.__onRemoved)
        self.OnSet += self.__ref(self.__onSet)
        self.add_field('From', 0, acceptIntGreaterThanOrEqualTo(0), str, '')
        self.add_field('To', 0, acceptIntGreaterThanOrEqualTo(0), str, '')
        self.add_field('Label', "", str, str, "")
        self.OnChangeListName = WeakEvent()
        self.__list_name = ""
        self.__order = []
        self.__lasts_order = None
        self.__building = False
        self.fits = []
        self.user_can_edit = True
    def set_list_name(self, x):
        if self.__list_name == x:
            return
        self.__list_name = x
        self.OnChangeListName(self, x)
    list_name = property(lambda self: self.__list_name, lambda self, x: self.set_list_name(x))
    def insert(self, i, entry, undoing=False):
        if self.__building:
            qubx.table.SimpleTable.insert(self, i, entry)
            self.__order.insert(i, entry['From'])
            self.__lasts_order = None
            self.fits.insert(i, None)
        elif ('From' in entry) and ('To' in entry):
            self.insert_selection(**entry)
        else:
            raise Exception('sorted table: use SelectionList.insert_selection instead')
    def insert_selection(self, From, To, Label="", **kw):
        ix = bisect.bisect_left(self.__order, From)
        kw['From'] = From
        kw['To'] = To
        kw['Label'] = Label
        qubx.table.SimpleTable.insert(self, ix, kw)
        self.__order.insert(ix, From)
        self.__lasts_order = None
        self.fits.insert(ix, None)
        return ix
    def insert_selections(self, sels):
        # if the final item in a sel tuple is a dict, that's used for keyword args. if one has kw, all must.
        if len(sels) and isinstance(sels[0][-1], dict):
            for sel in sels:
                self.insert_selection(*(sel[:-1]), **(sel[-1]))
        else:
            for sel in sels:
                self.insert_selection(*sel)
    def append_selection(self, From, To, Label="", **kw): # faster, if you're sure it belongs after the rest
        kw['From'] = From
        kw['To'] = To
        kw['Label'] = Label
        qubx.table.SimpleTable.append(self, kw)
        self.__order.append(From)
        self.__lasts_order = None
        self.fits.append(None)
        return self.size - 1        
    def selections_in(self, f, l, trimmed=False):
        if self.__lasts_order is None:
            self.__lasts_order = SortedCollection((self.get(i, 'To'), self.get(i, 'From'), self.get(i, 'Label'), i) for i in xrange(self.size))
        try:
            start = self.__lasts_order.index_ge((f,))
        except:
            return []

        sels = []
        for ix in xrange(start, self.size):
            To, From, Label, Index = self.__lasts_order[ix]
            if From > l:
                break
            if trimmed:
                if From < f: From = f
                if To > l: To = l
            sels.append((From, To, Label, Index))
        return sels
    def __onRemoved(self, index, undoing=False):
        del self.__order[index]
        self.__lasts_order = None
        del self.fits[index]
    def __onSet(self, index, field_name, val, prev, undoing=False):
        if field_name == 'From':
            qubx.table.run_later(self.__reinsert, index) # not reliable in a batch of arbitrary operations, or with  run_later==run_now
        else:
            self.__lasts_order = None
    def __reinsert(self, index):
        row = self.get_row(index)
        self.remove(index)
        kw = dict((field, row.__dict__[field]) for field in self.fields if not (field in ('Index', 'From', 'To')))
        self.insert_selection(row.From, row.To, **kw)
    def read_from_classic_tree(self, tree):
        self.clear()
        self.list_name = str(tree.data)
        for sel in qubx.tree.children(tree):
            if sel.name != 'Sels':
                self.insert_selection(sel.data[0], sel.data[1], sel.name)
    def to_classic_tree(self):
        tree = qubx.tree.Node('List')
        tree.data = self.list_name
        sel = qubx.tree.NullNode()
        for i in xrange(self.size):
            item = self.get_row(i)
            sel = tree.insert(item.Label, sel)
            sel.data = numpy.array([item.From, item.To], dtype='int32').reshape((1,2))
        return tree
    def read_from_tree(self, tree):
        self.list_name = str(tree.data)
        self.__building = True
        self.clear()
        fits_tree = tree.find('QUBX_FITS')
        if not fits_tree.isNull:
            tree.remove(fits_tree)
        qubx.table_tree.tree_to_table(tree, self)
        if not fits_tree.isNull:
            for i, fits_child in enumerate(qubx.tree.children(fits_tree)):
                if fits_child.child:
                    self.fits[i] = fits_child
        self.__building = False
    def to_tree(self):
        tree = qubx.table_tree.table_to_tree(self)
        tree.data = self.list_name
        have_fits = False
        for fit in self.fits:
            if not (fit is None):
                have_fits = True
                break
        if have_fits:
            fits_tree = tree['QUBX_FITS']
            fits_child = qubx.tree.NullNode()
            for fit in self.fits:
                if fit is None:
                    fits_child = fits_tree.insert('', fits_child)
                else:
                    fits_child = fits_tree.insertClone(fit, True, fits_child)
        return tree
    def from_text(self, txt):
        self.__building = True
        qubx.table.SimpleTable.from_text(self, txt, keep_fields=['From', 'To', 'Label'])
        self.__building = False
    def find_selection(self, From, To):
        for i in xrange(self.size):
            if (self[i,'From'] == From) and (self[i,'To'] == To):
                return i
        return -1

class SelectionLists(object):
    """

@ivar lists: list of L{SelectionList}
"""
    def __init__(self):
        self.__ref = Reffer()
        self.__index = self.__list = None
        self.lists = []
        self.OnSwitchList = WeakEvent()
        self.OnChangeListName = WeakEvent()
        self.OnChangeList = WeakEvent()
        self.OnSelect = WeakEvent()
        self.OnDoubleClick = WeakEvent()
    def set_index(self, x):
        if x >= len(self.lists):
            raise ValueError("index: %d >= len(lists): %d" % (x, len(self.lists)))
        if not (self.__list is None):
            for event in (self.__list.OnInsert, self.__list.OnRemoved, self.__list.OnSet):
                event -= self.__ref(self.__onEditList)
            self.__list.OnSelect -= self.__ref(self.__onSelect)
            self.__list.OnDoubleClick -= self.__ref(self.__onDoubleClick)
            self.__list.OnChangeListName -= self.__ref(self.__onChangeListName)
        if x < 0:
            self.__index = self.__list = None
        else:
            self.__index = x
            self.__list = self.lists[x]
            for event in (self.__list.OnInsert, self.__list.OnRemoved, self.__list.OnSet):
                event += self.__ref(self.__onEditList)
            self.__list.OnSelect += self.__ref(self.__onSelect)
            self.__list.OnDoubleClick += self.__ref(self.__onDoubleClick)
            self.__list.OnChangeListName += self.__ref(self.__onChangeListName)
        self.OnSwitchList(self, self.__index, self.__list)
    index = property(lambda self: self.__index, lambda self, x: self.set_index(x), doc="index of onscreen list")
    item = property(lambda self: self.__list, doc="onscreen L{SelectionList}")
    def __onEditList(self, *args):
        self.OnChangeList(self, self.__list)
    def __onSelect(self, index, field_name, sender):
        self.OnSelect(index, field_name, sender)
    def __onDoubleClick(self, index, field_name):
        self.OnDoubleClick(index, field_name)
    def __onChangeListName(self, lst, name):
        self.OnChangeListName(self, lst, name)
    def index_of(self, name):
        for i, lst in enumerate(self.lists):
            if lst.list_name == name:
                return i
        raise KeyError(name)
    def by_name(self, name):
        try:
            ix = self.index_of(name)
        except KeyError:
            lst = SelectionList()
            lst.list_name = name
            self.lists.append(lst)
            ix = len(self.lists) - 1
            if ix == 0:
                self.index = ix
        return self.lists[ix]
    def show_list(self, name):
        try:
            self.index = self.index_of(name)
        except KeyError:
            lst = SelectionList()
            lst.list_name = name
            self.lists.append(lst)
            self.index = len(self.lists) - 1
        return self.item
    def del_list(self, name_or_ix):
        try:
            ix = int(name_or_ix)
        except:
            ix = self.index_of(name_or_ix)
        if ix == self.__index:
            if len(self.lists) == 1:
                self.index = -1
            elif ix == (len(self.lists) - 1):
                self.index = self.index - 1
            else:
                self.index = self.index + 1
        del self.lists[ix]
        if self.__index > ix:
            self.__index -= 1
        elif self.__index == ix:
            if ix < len(self.lists):
                self.index = self.index
            else:
                self.index = self.__index - 1
    def read_from_classic_tree(self, lists_node):
        if lists_node:
            for i in reversed(xrange(len(self.lists))):
                self.del_list(i)
        for list_node in qubx.tree.children(lists_node, 'List'):
            self.show_list(list_node.name)
            self.item.read_from_classic_tree(list_node)
            if self.item.list_name == 'Segments':
                self.item.user_can_edit = False
        try:
            self.index = lists_node['ListIndex'].data[0]
        except:
            pass
    def to_classic_tree(self):
        tree = qubx.tree.Node('Lists')
        for lst in self.lists:
            tree.append(lst.to_classic_tree())
        tree['ListIndex'].data = self.index
        return tree
    def read_from_tree(self, lists_node):
        if lists_node:
            for i in reversed(xrange(len(self.lists))):
                self.del_list(i)
        for list_node in qubx.tree.children(lists_node, 'List'):
            self.show_list(list_node.name)
            self.item.read_from_tree(list_node)
            if self.item.list_name == 'Segments':
                self.item.user_can_edit = False
        try:
            self.index = lists_node['ListIndex'].data[0]
        except:
            pass
    def to_tree(self):
        tree = qubx.tree.Node('Lists_Express')
        for lst in self.lists:
            tree.append(lst.to_tree())
        tree['ListIndex'].data = self.index
        return tree
    def clone(self):
        other = SelectionLists()
        other.read_from_tree(self.to_tree())
        other.index = self.index
        return other
    def make_sub_list(self, name, offset, count, max_grow=0):
        src = self.item
        dest = self.show_list(name)
        dest.clear()
        for sel in src:
            first = sel.From + offset
            last = min(sel.To+max_grow, first + count - 1)
            if first <= last:
                dest.insert_selection(first, last, sel.Label)


class IdlUpdater(object):
    """Manages updates to idealization to make them undoable and to emit the OnChangeIdealization event; make sure to call done()."""
    def __init__(self, data, signal):
        self.data = data
        self.signal = signal
        self.defer_event = False
    def set_dwells(self, n, ff, ll, cc, aa=None, ss=None, event=False):
        if n <= 0:
            return
        idl = self.data.ideal[self.signal].idl
        idlmeta = self.data.ideal[self.signal].seg[self.data.segmentation.index_at(ff[0])]
        def make_set(ff, ll, cc, aa, ss, ev):
            aaa = aa if (aa is None) else aa[:]
            sss = ss if (ss is None) else ss[:]
            def do():
                idl.set_dwells(len(ff), ff, ll, cc)
                if not ((idlmeta.amp is None) or (aaa is None)) and (len(idlmeta.amp) >= len(aaa)):
                    idlmeta.amp[:len(aaa)] = aaa
                    if not (sss is None):
                        idlmeta.std[:len(sss)] = sss
                elif not (aaa is None):
                    idlmeta.amp = aaa
                    if not (sss is None):
                        idlmeta.std = sss
                    else:
                        idlmeta.std = [0.0] * len(aaa)
                if ev:
                    self.data.OnChangeIdealization(self.signal)
            return do
            
        uff, ull, ucc = idl.get_dwells_and_gaps(ff[0], ll[-1])
        rff, rll, rcc = [numpy.array(zz, copy=True) for zz in (ff, ll, cc)]
        raa = aa if (aa is None) else aa[:]
        rss = ss if (ss is None) else ss[:]
        self.data.undoIdl.push_undo(make_set(uff, ull, ucc, idlmeta.amp, idlmeta.std, True),
                                    make_set(rff, rll, rcc, raa, rss, True))
        make_set(ff, ll, cc, aa, ss, event)()
        if not event:
            self.defer_event = True
    def done(self):
        self.data.undoIdl.seal_undo('Idealize')
        if self.defer_event:
            self.data.OnChangeIdealization(self.signal)


STIM_TYPES = (0, 1, 2)
(STIM_TYPE_CONST, STIM_TYPE_SIGNAL, STIM_TYPE_EXPR) = STIM_TYPES
STIM_LABELS = ('Constant', 'Signal', 'Expression')

def acceptStimType(s):
    try:
        return STIM_LABELS.index(s)
    except:
        raise FieldAcceptException(s, 'must be one of %s' % STIM_LABELS)
def formatStimType(x):
    return STIM_LABELS[x]


class StimulusRecord(Anon):
    """
Describes how to find and idealize the time course of a model variable.
It can be constant, an analog signal, or an expression in terms of segment index, time since seg-start, and named analog signals.  Read-only please.  You are more likely to encounter a L{qubx.table.TableRow} with the same fields.

@ivar Name: the model variable e.g. 'Voltage'
@ivar Type: STIM_TYPE_CONST, STIM_TYPE_SIGNAL, or STIM_TYPE_EXPR
@ivar Signal_index: index in dataview.signals, if stim_type == STIM_TYPE_SIGNAL
@ivar Expr: f(seg, t[, signalname[0], signalname[1], ...]) -> float, if stim_type == STIM_TYPE_EXPR
@ivar Latency: [ms] milliseconds of delay to add
@ivar Shaping: name of a shaping plugin
@ivar Known_amps: list of known stimulus levels for idealization
@ivar Add_deltas: whether to add more stimulus levels when idealizing
@ivar Delta_amp: minimum difference between added levels
@ivar Min_dur: [ms] events any shorter are subsumed by their neighbor
"""
    def __init__(self, name, stim_type=STIM_TYPE_CONST, signal_index=None, expr="0", latency_ms=0.0,
                 shaping="", known_amps=[0.0], add_deltas=True, delta=0.5, min_dur_ms=1.0, expr_index=None, active=True):
        Anon.__init__(self, Name=name, Type=stim_type, Signal_index=signal_index, Expr=expr, Latency=latency_ms,
                      Shaping=shaping, Known_amps=known_amps, Add_deltas=add_deltas, Delta_amp=delta, Min_dur=min_dur_ms)

class QubData(object):
    """Base class for an open data file.

    @ivar sampling: interval between data points, in seconds
    @ivar segmentation: L{Segmentation}
    @ivar seg: list of objects to hold per-segment attributes
    @ivar path: folder/filename or ""
    @ivar exclusion: L{Exclusion} -- idealization of erased regions
    @ivar analog: list of L{QubData_Analog} for each signal
    @ivar overlays: list of L{SigOverlays} (or None) for each signal
    @ivar baseline: list of L{SigBaseline} for each signal
    @ivar fits: list of L{SigIdl<IdlFits>} for each signal
    @ivar ideal: list of L{SigIdl<Idealization>} for each signal
    @ivar qsf: L{qubx.tree.Node} with metadata, saved as <whatever>.qsf
    @ivar constants: L{qubx.table.SimpleTable} with Name, Value
    @ivar signals: L{qubx.table.SimpleTable} with Name, Units, Scale, Offset, Filter, Filter Freq
    @ivar stimuli: L{qubx.table.SimpleTable} with the same fields as L{StimulusRecord}.
    @ivar segments: L{qubx.table.SimpleTable} of stats etc; auto-adds fields
    @ivar lists: L{SelectionLists}
    @ivar list: currently showing L{SelectionList}
    @ivar notes: comment string for end-use
    @ivar trials: L{qubx.trialset.TrialSets}
    #ivar store_fit_curves: True to save qsf['FitCurves'], default False to save space
    @ivar OnChangePath: L{WeakEvent}(QubData, path or "") when filename or path change
    @ivar OnChangeField: L{WeakEvent}(QubData, 'Name' or 'In Folder') when filename or path change, for L{qubx.table.ObjectTable}
    @ivar OnChangeSampling: L{WeakEvent}(QubData, sampling) when the sampling rate changes
    @ivar OnChangeSamples: L{WeakEvent}(QubData, signal_index or -1, *reserved) when sampled data changes
    @ivar OnChangeIdealization: L{WeakEvent}(QubData, signal_index or -1, *reserved) when idealized data changes
    @ivar OnChangeFits: L{WeakEvent}(QubData, signal_index or -1, *reserved) when fit curve changes
    @ivar OnChangeBaselineNodes: L{WeakEvent}(QubData, signal_index or -1)
    @ivar OnChangeOverlaySource: L{WeakEvent}(QubData, signal_index)
    @ivar OnChangeOverlays: L{WeakEvent}(QubData, signal_index)
    @ivar OnSaving: L{WeakEvent}(QubData, qsf) when qsf (L{qubx.tree.Node}) needs updating prior to save
"""
    def __init__(self):
        self.__ref = Reffer()
        self.__sampling = 1.0
        self.scaling = 1.0
        self.OnChangePath = WeakEvent() # (QubData, path or "")
        self.OnChangeField = WeakEvent() # (QubData, 'Name' or 'In Folder') for ObjectTable
        self.OnChangeSampling = WeakEvent() # {QubData, sampling)
        self.OnChangeSamples = WeakEvent() # (QubData, signal_index, *reserved)
        self.OnChangeIdealization = WeakEvent() # (QubData, signal_index, *reserved)
        self.OnChangeFits = WeakEvent() # (QubData, signal_index, *reserved)
        self.OnChangeBaselineNodes = WeakEvent()
        self.OnChangeOverlaySource = WeakEvent()
        self.OnChangeOverlays = WeakEvent()
        self.OnChangeOverlays += self.OnChangeSamples
        self.OnSaving = WeakEvent()
        self.constants = SimpleTable('Constants', global_name='QubX.Data.file.constants')
        self.constants.add_field('Name', '', acceptString, str, '')
        self.constants.add_field('Value', 1.0, acceptFloat, str, '')
        self.signals = SimpleTable('Signals', global_name='QubX.Data.file.signals')
        self.signals.add_field('Name', '', acceptNothing, str, '')
        self.signals.add_field('Units', '', acceptNothing, str, '')
        self.signals.add_field('Scale', 1.0, acceptFloat, str, '')
        self.signals.add_field('Offset', 0.0, acceptFloat, str, '')
        self.signals.add_field('Filter', False, acceptBool, str, '')
        self.signals.add_field('Filter Freq', 1.0, acceptFloat, '%.3g', 'kHz')
        self.signals.add_field('Computed as', '', str, str, '')
        self.signals.OnInsert += self.__ref(self.__onInsertSignal)
        self.signals.OnRemoving += self.__ref(self.__onRemovingSignal)
        self.signals.OnSet += self.__ref(self.__onSetSignal)
        self.analog = []
        self.overlays = []
        self.analog_sources = weakref.WeakValueDictionary()
        self.baseline = []
        self.fits = []
        self.store_fit_curves = False
        self.ideal = []
        self.stim_expr_ideal = collections.defaultdict(lambda: SigIdl(self.sampling, self.segmentation))
        self.segmentation = Segmentation()
        self.segmentation.OnAdd += self.__ref(self.__onAddSeg)
        self.segmentation.OnClear += self.__ref(self.__onClearSegs)
        self.segments = SimpleTable('Segments', auto_add_fields=True, global_name='QubX.Data.file.segments')
        self.segments.add_field('SampleCount', 0, acceptNothing, str, '')
        self.segments.add_field('Start', 0.0, acceptNothing, '%.5g', 'ms', independent=True)
        self.segments.add_field('Duration', 0.0, acceptNothing, '%.5g', 'ms')
        self.segments.add_field('Label', '', acceptString, str, '') # required for extract:list to copy List table to Segments table
        self.segments.OnSet += self.__ref(self.__onSetSegments)
        self.seg_baseline = []
        self.seg_bl_undo = []
        self.seg = [] # per-seg attributes
        self.exclusion = Exclusion(self.__sampling, self.segmentation)
        self.exclusion.OnChange += self.__ref(self.__onExclusionChange)
        self.stimuli = SimpleTable('Stimuli', global_name='QubX.Data.file.stimuli')
        self.stimuli.add_field('Name', '', acceptNothing, str, '')
        self.stimuli.add_field('Type', STIM_TYPE_CONST, acceptStimType, formatStimType, '', STIM_TYPES)
        self.stimuli.add_field('Signal_index', 0, acceptIntGreaterThan(-1), str, '')
        acceptExpr = acceptF(static=[], custom=True)
        self.stimuli.add_field('Expr', acceptExpr('0'), acceptExpr, str, '')
        self.stimuli.add_field('Latency', 0.0, acceptFloat, '%.3g', 'ms')
        self.stimuli.add_field('Shaping', '', str, str, '')
        self.stimuli.add_field('Known_amps', [], acceptFloatList(), formatList('%.3g'), '')
        self.stimuli.add_field('Add_deltas', True, acceptBool, str, '')
        self.stimuli.add_field('Delta_amp', 0.5, acceptFloat, '%.3g', '')
        self.stimuli.add_field('Min_dur', 0.0, acceptFloatGreaterThanOrEqualTo(0.0), '%.3g', 'ms')
        self.stimuli.OnInsert += self.__ref(self.__onInsertStimulus)
        self.stimuli.OnSet += self.__ref(self.__onSetStimulus)
        self.lists = SelectionLists()
        self.path = ""
        self.qsf = qubx.tree.Node('SessionFile')
        self.group = 1
        self.notes = ""
        self.trials = qubx.trialset.TrialSets()
        self.undoIdl = qubx.undo.UndoStack(depth=2)
        self.__undoIdlSignal = None
    def dispose(self):
        for tbl in (self.signals, self.segments, self.stimuli, self.constants):
            tbl.dispose()
        self.__ref.clear()
    list = property(lambda self: self.lists.item)
    def set_sampling(self, x):
        self.__sampling = x
        self.exclusion.sampling = x
        for idl in self.ideal:
            if idl:
                idl.sampling = x
        for fit in self.fits:
            if fit:
                fit.sampling = x
        self.OnChangeSampling(self, x)
    sampling = property(lambda self: self.__sampling, lambda self, x: self.set_sampling(x), "delta t (seconds)")
    def get_stimulus(self, name, def_latency=0.0, def_known_amps=[0.0], def_add_deltas=True, def_delta=0.5, def_min_dur=1.0, def_value=1.0):
        """Returns the L{StimulusRecord}, creating with defaults if needed."""
        try:
            i = self.stimuli.index(name)
            return self.stimuli.get_row_by_name(name)
        except KeyError:
            sig_names = [self.signals.get(i, 'Name') for i in xrange(self.signals.size)]
            try:
                signal_index = sig_names.index(name)
                stim_type = STIM_TYPE_SIGNAL
            except:
                signal_index = 0
                stim_type = STIM_TYPE_CONST
            if stim_type == STIM_TYPE_CONST:
                const_names = [self.constants.get(i, 'Name') for i in xrange(self.constants.size)]
                if not (name in const_names):
                    self.constants.append({'Name':name, 'Value':def_value})
            self.stimuli.append({'Name' : name,
                                 'Type' : stim_type,
                                 'Signal_index' : signal_index,
                                 'Latency' : def_latency,
                                 'Known_amps' : def_known_amps,
                                 'Add_deltas' : def_add_deltas,
                                 'Delta_amp' : def_delta,
                                 'Min_dur' : def_min_dur})
            return self.stimuli.get_row_by_name(name)
    def get_stimulus_at(self, name, f):
        """Returns the value of a stimulus variable at the sample index f."""
        stim = self.get_stimulus(name)
        try:
            stim.Type
        except:
            print stim, stim.fields
        if stim.Type == STIM_TYPE_CONST:
            for i in xrange(self.constants.size):
                if name == self.constants.get(i, 'Name'):
                    return self.constants.get(i, 'Value')
        elif stim.Type == STIM_TYPE_SIGNAL:
            ff, ll, cc, dd, aa = self.get_stimulus_idl(stim.Name, f, f)
            if len(cc) and (cc[0] >= 0):
                return aa[cc[0]]
            else:
                return self.analog[stim.Signal_index].read_overlaid(self, stim.Signal_index, f, f, int(round(stim.Latency*1e-3/self.sampling)))[0]
        elif stim.Type == STIM_TYPE_EXPR:
            ff, ll, cc, dd, aa = self.get_stimulus_idl(stim.Name, f, f)
            return aa[cc[0]]
    def get_stimulus_idl(self, name, f, l):
        """Returns the idealized stimulus (firsts, lasts, classes, durations, amps) for named variable in sample range (f, l)."""
        stim = self.get_stimulus(name)
        if stim.Type == STIM_TYPE_CONST:
            return (numpy.array([f], dtype='int32'), 
                    numpy.array([l], dtype='int32'), 
                    numpy.array([0], dtype='int32'), 
                    numpy.array([self.sampling*1e3*(l-f+1)], dtype='float32'),
                    numpy.array([self.get_stimulus_at(name, 0)], dtype='float64'))
        elif stim.Type == STIM_TYPE_SIGNAL:
            ideal = self.ideal[stim.Signal_index]
        elif stim.Type == STIM_TYPE_EXPR:
            ideal = self.stim_expr_ideal[name]
        firsts, lasts, classes, durations = ideal.idl.get_dwells(f, l, True, True)
        amp_src = ideal.seg[self.segmentation.index_at(f)].amp
        if (amp_src is None) or (len(amp_src) == 0):
            amp_src = []
        amps = numpy.array(amp_src, dtype='float64')
        return firsts, lasts, classes, durations, amps
    def read_session(self, qsf, progressf=lambda pct:True):
        try:
            self.qsf = qsf
            if qsf.find('Signals'):
                tree_to_table(qsf['Signals'], self.signals)
            else: # classic QuB:
                filters, freqs = [qsf.find('Display').find(x).data for x in ['FilterOn', 'FilterFreq']]
                for c, chan in enumerate(qubx.tree.children(qsf.find('DataChannels'), 'Channel')):
                    if c >= self.signals.size:
                        break
                    self.signals.set(c, 'Name', str(chan['Name'].data))
                    self.signals.set(c, 'Units', str(chan['Units'].data))
                    self.signals.set(c, 'Scale', 1.0 / (chan['Scaling'].data and chan['Scaling'].data[0] or 1.0))
                    self.signals.set(c, 'Offset', chan['Offset'].data and chan['Offset'].data[0] or 0.0)
                    if c < len(filters):
                        self.signals.set(c, 'Filter', bool(filters[c]))
                    if c < len(freqs):
                        self.signals.set(c, 'Filter Freq', 1e-3*freqs[c])
            if qsf.find('Stimuli'):
                tree_to_table(qsf['Stimuli'], self.stimuli)
                tree_to_table(qsf['Constants'], self.constants)
            else: # classic QuB:
                for var in qubx.tree.children(qsf.find('ExpCond')):
                    ind = var.find('ChannelIndex').data
                    if ind and (ind[0] < self.signals.size):
                        stim = self.get_stimulus(var.name)
                        self.stimuli.set(stim.Index, 'Type', STIM_TYPE_SIGNAL)
                        self.stimuli.set(stim.Index, 'Signal_index', ind[0])
                    elif var.data:
                        try:
                            self.constants.set(self.constants.index(var.name), 'Value', var.data[0])
                        except:
                            self.constants.append({'Name' : var.name, 'Value' : var.data[0]})
            if qsf.find('Lists_Express'):
                self.lists.read_from_tree(qsf.find('Lists_Express'))
            else:
                self.lists.read_from_classic_tree(qsf.find('Lists'))
            ReadSessionIdl(self, qsf, addsegs=(not self.segmentation.segments), progressf=progressf)
            ReadSessionExcl(self, qsf, progressf=progressf)
            ReadSessionBaseline(self, qsf, progressf=progressf)
            ReadSessionTrials(self, qsf, progressf=progressf)
            ReadSessionFits(self, qsf, progressf=progressf)
            ReadSessionOverlays(self, qsf, progressf=progressf)
            if qsf.find('Segments'):
                tree_to_table(qsf['Segments'], self.segments)
            if qsf.find('DataGroup').data:
                self.group = qsf['DataGroup'].data[0]
            self.notes = str(qsf.find('Comment').data)
        except:
            traceback.print_exc()
    def saveAs(self, path=None, progressf=lambda pct: True):
        """Writes the data according to file extension (default txt)."""
        segs_of_signal = [self.get_segmentation_file(signal=s, baseline_nodes=False) for s in xrange(self.signals.size)]
        ext = os.path.splitext(path)[1].lower()
        if ext == '':
            ext = '.txt'
            path = path + ext
        tables = self.get_save_tables()
        constants, segments, signals, stimuli = tables[:4]
        signals = tables[2] = signals.clone()
        for s in xrange(signals.size):
            signals.set(s, 'Scale', 1.0)
            signals.set(s, 'Offset', 0.0)
        Save(path, self.qsf, segs_of_signal, self.lists, progressf, *tables)
    def pre_close(self):
        self.save_session()
    def save_session(self):
        if os.path.split(self.path)[0]:
            segs_of_signal = [self.get_segmentation_file(signal=s) for s in xrange(self.signals.size)]
            qsf = BuildSession(self.qsf, segs_of_signal, self.lists, *self.get_save_tables())
            try:
                BuildSessionBaseline(qsf, self.baseline)
                BuildSessionTrials(qsf, self.trials)
                qsf['Comment'].data = self.notes
                if self.store_fit_curves:
                    BuildSessionFits(qsf, self, segs_of_signal)
                BuildSessionOverlays(qsf, self, segs_of_signal)
                qsf.saveAs(os.path.splitext(self.path)[0] + '.qsf', as_copy=True)
            except:
                traceback.print_exc()
    def get_save_tables(self):
        tables = [self.constants, self.segments, self.signals, self.stimuli]
        self.OnSaving(self, tables)
        return tables
    def set_analog(self, i, analog):
        self.analog[i] = analog
        self.analog[i].sampling = self.sampling
        self.analog_sources[SafeName(self.signals[i, 'Name'])] = analog
        self.OnChangeSamples(self, i)
    def get_overlays(self, i):
        if self.overlays[i] is None:
            self.overlays[i] = SigOverlays(self.sampling, self.segmentation)
            self.overlays[i].OnChangeSource += self.__ref(bind(self.OnChangeOverlaySource, self, i))
            self.overlays[i].OnChangeOverlays += self.__ref(bind(self.OnChangeOverlays, self, i))
        return self.overlays[i]
    def accept_computed(self, expr):
        return parse_computed_signal(expr, self.analog_sources.keys())
    def __onInsertSignal(self, i, undoing):
        self.analog.insert(i, None)
        self.overlays.insert(i, None)
        self.baseline.insert(i, SigBaseline(self.segmentation))
        self.baseline[i].OnChange += self.__ref(self.__onChangeBaseline)
        self.fits.insert(i, SigIdl(self.sampling, self.segmentation, IdlFits))
        self.ideal.insert(i, SigIdl(self.sampling, self.segmentation))
        qubx.table.run_later(self.__connect_stim_signal_if, i)
        name = self.signals[i, 'Name']
        baseline_field = 'Baseline offset %s' % name
        if baseline_field in self.segments.fields:
            seg_baseline = [self.segments[s, baseline_field] for s in len(self.segmentation.segments)]
        else:
            seg_baseline = [None for s in xrange(len(self.segmentation.segments))]
        self.seg_baseline.insert(i, seg_baseline)
        self.seg_bl_undo.insert(i, [UndoStack() for s in xrange(len(self.segmentation.segments))])
        expr = self.signals[i, 'Computed as']
        if name and expr:
            try:
                self.set_analog(i, QubData_Analog_Computed(self.accept_computed(expr), self.analog_sources, self, i))
            except NameError, ne:
                print 'Error computing signal %s=%s:\n%s' % (name, expr, ne)
        else:
            self.signals.user_can_remove[i] = False
    def __connect_stim_signal_if(self, sig_ix):
        for stim_ix in xrange(self.stimuli.size):
            stim = self.stimuli.get_row(stim_ix)
            if (stim.Type == STIM_TYPE_CONST) and (stim.Name == self.signals.get(sig_ix, 'Name')):
                self.stimuli.set(stim_ix, 'Type', STIM_TYPE_SIGNAL)
                self.stimuli.set(stim_ix, 'Signal_index', sig_ix)
                break
    def __onRemovingSignal(self, i, undoing):
        del self.analog[i]
        del self.overlays[i]
        self.baseline[i].OnChange -= self.__ref(self.__onChangeBaseline)
        del self.baseline[i]
        del self.fits[i]
        del self.ideal[i]
        del self.seg_baseline[i]
        del self.seg_bl_undo[i]
        for ix in xrange(self.stimuli.size):
            if ((self.stimuli.get(ix, 'Type') == STIM_TYPE_SIGNAL) and
                (self.stimuli.get(ix, 'Signal_index') == i)):
                self.stimuli.set(ix, 'Type', STIM_TYPE_CONST)
        self.OnChangeSamples(self, None)
        self.OnChangeIdealization(self, None)
    def __onSetSignal(self, i, field, val, prev, undoing):
        if (field == 'Filter') or ((field == 'Filter Freq') and self.signals.get(i, 'Filter')):
            self.OnChangeSamples(self, i)
        elif field == 'Name':
            self.__connect_stim_signal_if(i)
            if self.analog[i]:
                if SafeName(prev) in self.analog_sources:
                    del self.analog_sources[SafeName(prev)]
                self.analog_sources[SafeName(val)] = self.analog[i]
        elif field == 'Scale':
            bl_name = 'Baseline offset %s' % self.signals[i, 'Name']
            if bl_name in self.segments.fields:
                offset = self.signals[i, 'Offset']
                raw = lambda v: (v - offset) / prev
                cooked = lambda v: v * val + offset
                for s in xrange(self.segments.size):
                    self.segments[s, bl_name] = cooked(raw(self.segments[s, bl_name]))
            self.OnChangeSamples(self, i)
        elif field == 'Offset':
            bl_name = 'Baseline offset %s' % self.signals[i, 'Name']
            if bl_name in self.segments.fields:
                dy = val - prev
                for s in xrange(self.segments.size):
                    self.segments[s, bl_name] = dy + self.segments[s, bl_name]
            self.OnChangeSamples(self, i)
        elif field == 'Computed as':
            if (not self.analog[i]) or hasattr(self.analog[i], 'func'):
                self.signals.user_can_remove[i] = True
                try:
                    func = self.accept_computed(val)
                    if self.analog[i]:
                        self.analog[i].func = func
                        self.OnChangeSamples(self, i)
                    else:
                        self.set_analog(i, QubData_Analog_Computed(func, self.analog_sources, self, i))
                except NameError, ne:
                    qubx.pyenvGTK.show_message(str(ne))
            elif val:
                qubx.pyenv.env.call_later(self.signals.set, i, field, '')
                
            
    def __onSetSegments(self, i, field, val, prev, undoing):
        match = re.match(r"Baseline offset (.+)", field)
        if match:
            try:
                sig_ix = self.signals.index(match.group(1))
                self.seg_baseline[sig_ix][i] = val
                if not undoing:
                    self.seg_bl_undo[sig_ix][i].push_undo(lambda: self.segments.set(i, field, prev, undoing=True),
                                                          lambda: self.segments.set(i, field, val, undoing=True))
                    self.seg_bl_undo[sig_ix][i].seal_undo('Set segment baseline')
                self.OnChangeSamples(self, sig_ix)
            except:
                traceback.print_exc()
    def __onAddSeg(self, segm, f, l, start):
        self.seg.append(Anon(f=f, l=l, ix=len(self.seg)))
        self.segments.append({'SampleCount' : (l - f + 1),
                              'Duration'    : (l - f + 1) * 1e3 * self.sampling,
                              'Start'       : start})
        self.segments.user_can_remove[len(self.segments) - 1] = False
        for sig in self.seg_baseline:
            sig.append(0.0)
        for sig in self.seg_bl_undo:
            sig.append(UndoStack())
    def __onClearSegs(self, segm):
        self.seg = []
        self.segments.clear()
        self.seg_baseline = [ [] for i in xrange(len(self.signals)) ]
        self.seg_bl_undo = [ [] for i in xrange(len(self.signals)) ]
    def __onChangeBaseline(self):
        self.OnChangeBaselineNodes(self, -1) # pass thru signal index?
        self.OnChangeSamples(self, -1)
    def __onExclusionChange(self, excl):
        self.OnChangeSamples(self, -1)
        self.OnChangeIdealization(self, -1)
        self.OnChangeFits(self, -1)
    def __onInsertStimulus(self, i, undoing):
        if self.stimuli.get(i, 'Type') == STIM_TYPE_CONST:
            self.__ensure_constant(self.stimuli.get(i, 'Name'))
    def __ensure_constant(self, name, default=1.0):
        try:
            self.constants.index(name)
        except KeyError:
            self.constants.append({'Name' : name, 'Value' : default})
    def __onSetStimulus(self, i, field, val, prev, undoing):
        stim = self.stimuli.get_row(i)
        if field == 'Type':
            if val == STIM_TYPE_CONST:
                self.__ensure_constant(stim.Name)
        changed = []
        sig = (stim.Type == STIM_TYPE_SIGNAL) and stim.Signal_index or None
        if field == 'Latency':
            if not (sig is None):
                changed.append(sig)
        elif stim.Latency != 0.0:
            if field == 'Type':
                if (val == STIM_TYPE_SIGNAL) or (prev == STIM_TYPE_SIGNAL):
                    changed.append(stim.Signal_index)
            elif not (sig is None):
                if field == 'Signal_index':
                    changed.append(val)
                    changed.append(prev)
                else:
                    changed.append(sig)
        for sig in changed:
            self.OnChangeSamples(self, sig)
    def get_segmentation_file(self, signal=0, latency=None, baseline_nodes=True):
        """Returns a list of L{SourceSeg} corresponding to the whole file."""
        left = right = None
        return self.get_segmentation(0, len(self.segmentation.segments)-1, left, right, signal=signal, latency=latency, baseline_nodes=baseline_nodes)
    def get_segmentation(self, first_seg=None, last_seg=None, left=None, right=None, signal=0, latency=None, baseline_nodes=True):
        """Returns a list of all L{SourceSeg}s, modified by args.

        @param first_seg:
        @param last_seg:
        @param left: time offset from start-of-each-seg
        @param right: time offset from start-of-each-seg
        @param signal: index in file.signals
        @param latency: number of samples to shift the data rightwards (same f, l, and idealization)
"""
        lat = 0 if (latency is None) else latency
        if lat is None:
            if self.stimuli:
                for stim in self.stimuli.all_rows():
                    if (stim.Type == STIM_TYPE_SIGNAL) and (signal == stim.Signal_index):
                        lat = int(round(stim.Latency*1e-3/self.sampling))
                        break
        filter_Hz = 1e3*(self.signals.get(signal, 'Filter') and self.signals.get(signal, 'Filter Freq') or 0.0)
        segf, segl, l, r = first_seg, last_seg, left, right
        if segf is None:
            segf = 0
        if segl is None:
            segl = len(self.segmentation.segments) - 1
        if l is None:
            l = 0.0
        if r is None:
            r = 1e20
        segs = []
        for iseg in xrange(segf, segl+1):
            first, last, n = self.segmentation.segments[iseg]
            group = self.segments[iseg, 'Group']
            dt = self.sampling
            first_in = max(first, first + int(round(l/dt)))
            last_in = min(last, first + int(round(r/dt)))
            if first_in > last_in:
                continue
            start_seg = self.segmentation.starts[iseg]
            seg = SourceSeg(self, signal, iseg, first, first_in, last_in,
                            latency=lat, filter_Hz=filter_Hz,
                            start=start_seg + (first_in - first)*1e3*dt,
                            baseline_nodes=baseline_nodes, group=group)
            for ef, el, ec in izip(*self.exclusion.idl.get_dwells(first_in, last_in, True)):
                seg.chunks.append(SourceChunk(self, signal, ef, el, ((ec!=0) and (el-ef+1) or 0),
                                              latency=lat, filter_Hz=filter_Hz,
                                              start=start_seg + (ef - first)*1e3*dt,
                                              baseline_nodes=baseline_nodes,
                                              group=group))
            segs.append(seg)
        return segs
    def get_segmentation_indexed(self, first, last, signal=0, latency=None, baseline_nodes=True, group=None):
        filter_Hz = 1e3*(self.signals.get(signal, 'Filter') and self.signals.get(signal, 'Filter Freq') or 0.0)
        if latency is None:
            lat = 0
            for stim in self.stimuli.all_rows():
                if (stim.Type == STIM_TYPE_SIGNAL) and (signal == stim.Signal_index):
                    lat = int(round(stim.Latency*1e-3/self.sampling))
                    break
        else:
            lat = latency

        segs = []
        for iseg, f, l in self.segmentation.split(first, last):
            g = self.segments[iseg, 'Group'] if (group is None) else group
            seg_f = self.segmentation.segments[iseg][0]
            seg_start = self.segmentation.starts[iseg]
            seg = SourceSeg(self, signal, iseg, seg_f, seg_f + f, seg_f + l,
                            latency=lat, filter_Hz=filter_Hz,
                            start=seg_start + f*1e3*self.sampling,
                            baseline_nodes=baseline_nodes,
                            group=g)
            for ef, el, ec in izip(*self.exclusion.idl.get_dwells(seg_f+f, seg_f+l, True)):
                seg.chunks.append(SourceChunk(self, signal, ef, el, ((ec!=0) and (el-ef+1) or 0),
                                              latency=lat, filter_Hz=filter_Hz,
                                              start=seg_start + (ef - seg_f)*1e3*self.sampling,
                                              baseline_nodes=baseline_nodes,
                                              group=g))
            segs.append(seg)
        return segs
    def get_segmentation_list(self, list_name=None, signal=0, baseline_nodes=True, group=None):
        segs = []
        lst = self.list if (list_name is None) else self.lists.by_name(list_name)
        if lst:
            for i in xrange(lst.size):
                sel = lst.get_row(i)
                if (group is None) or (sel.Group == group):
                    segs.extend(self.get_segmentation_indexed(sel.From, sel.To, signal=signal, baseline_nodes=baseline_nodes, group=sel.Group))
        return segs
    def update_idl(self, signal):
        if signal != self.__undoIdlSignal:
            self.__undoIdlSignal = signal
            self.undoIdl.clear()
        return IdlUpdater(self, signal)


def get_segmentation_copy(segmentation, signal=0):
    if isinstance(segmentation, collections.Iterable):
        lists = [seg.file.get_segmentation_indexed(seg.f, seg.l, signal) for seg in segmentation]
        return [lst[0] for lst in lists if len(lst)]
    else:
        return get_segmentation_copy([segmentation], signal)[0]


class SourceChunk(Anon):
    """Describes one contiguous slice of data from the data source.

    @ivar file: L{qubx.data.QubData}
    @ivar signal: index of source signal
    @ivar f: first data point
    @ivar l: last data point
    @ivar n: number of data points
    @ivar start: in ms
    @ivar included: False if excluded
    @ivar latency: number of samples to shift the data rightwards (ignored for idealization)
    @ivar filter_Hz: low-pass filter frequency, or None
    @ivar sampling: sampling interval (sec)
    @ivar idealized: True if segment has (starts with) idealized dwells
    @ivar baseline_nodes: True to subtract baseline nodes
    @ivar group: Group column from List or Segments
"""
    def __init__(self, file=None, signal=0, f=0, l=-1, included=True, latency=0, start=-1, sampling=None, baseline_nodes=True, group=0, **kw):
        if not ('n' in kw):
            kw['n'] = l-f+1
        if not ('idealized' in kw):
            kw['idealized'] = (0 <= signal < len(file.ideal)) and (0 <= file.ideal[signal].idl[f])
        Anon.__init__(self, file=file, signal=signal, f=f, l=l, included=included, latency=latency, start=start,
                      sampling=sampling or file.sampling, baseline_nodes=baseline_nodes, group=group, **kw)
    def get_actual_bounds(self, first_offset=None, last_offset=None):
        f, l, start = self.f, self.l, self.start
        if not (first_offset is None):
            f += first_offset
            start += first_offset * self.sampling
        if not (last_offset is None):
            l = min(self.l, self.f + last_offset)
        return f, l, start
    def get_samples(self, first_offset=None, last_offset=None):
        f, l, start = self.get_actual_bounds(first_offset, last_offset)
        if self.filter_Hz:
            return get_file_samples_filtered(self.file, self.signal, f, l, self.included, self.latency,
                                             self.filter_Hz, start=start, baseline_nodes=self.baseline_nodes, group=self.group)
        else:
            return get_file_samples(self.file, self.signal, f, l, self.included, self.latency, 1, start, baseline_nodes=self.baseline_nodes, group=self.group)
    def get_idealization(self, signal=None, mark_excluded=False, get_fragments=False, get_durations=False,
                         get_amps=False, get_stds=False, first_offset=0, last_offset=None):
        """Returns the dwells in chunk as numpy arrays

        @param signal: signal index, or None to use the chunk.signal
        @param mark_excluded: if True, replaces excluded dwells with class -1 (or -2?)
        @param get_fragments: if False, omits first/last dwells if incomplete
        @param get_durations: if True, returns durations
        @param get_amps: if True: returns amps
        @param get_stds: if True: returns stds
        @param first_offset: (optional) to narrow the desired time range
        @param last_offset: (optional)
        @return: firsts, lasts, classes[, durations, amps, stds]
"""
        f = self.f + first_offset
        l = self.l if (last_offset is None) else min(self.l, self.f + last_offset)
        sig = signal
        if (sig is None) or not (0 <= sig < self.file.signals.size):
            sig = self.signal
        if not (0 <= sig < self.file.signals.size):
            rtn = [ [], [], [] ]
            if get_durations:
                rtn.append( [] )
            if get_amps:
                rtn.append( [] )
            if get_stds:
                rtn.append( [] )
        else:
            if mark_excluded:
                rtn = list(self.file.ideal[sig].idl.get_dwells_excluding(self.file.exclusion.idl, 0,
                                                                         f, l, get_durations))
            else:
                rtn = list(self.file.ideal[sig].idl.get_dwells(f, l, get_fragments, get_durations))
            if get_amps:
                rtn.append(self.file.ideal[sig].seg[ self.file.segmentation.index_at(f) ].amp)
            if get_stds:
                rtn.append(self.file.ideal[sig].seg[ self.file.segmentation.index_at(f) ].std)
        return rtn


class SourceSeg(SourceChunk):
    """Describes one segment from the data source.

    @ivar index: index of segment in file
    @ivar offset: sample index offset of source segment in file
    @ivar chunks: list of L{SourceChunk}
"""
    def __init__(self, file, signal=0, index=0, offset=0, f=0, l=-1, chunks=None, **kw):
        if not chunks:
            chunks=[]
        SourceChunk.__init__(self, file=file, signal=signal, f=f, l=l, index=index, offset=offset, chunks=chunks, **kw)


def get_chunk_samples(chunk):
    """Returns a L{SourceChunk} with samples and sampling."""
    return chunk.get_samples()

def get_file_samples(file, signal, first, last, included, latency=0, skip=1, start=-1, baseline_nodes=True, group=None):
    """Returns a L{SourceChunk} with samples and sampling."""
    g = 0
    if (first > last):
        samples = numpy.zeros(shape=(0,))
    elif (0 <= signal < len(file.analog)) and file.analog[signal]:
        iseg = file.segmentation.index_at(first)
        samples = file.analog[signal].read_overlaid(file, signal, first, last, latency, skip)
        bs_off = file.seg_baseline[signal][iseg]
        if bs_off:
            samples -= bs_off
        if baseline_nodes and len(file.baseline[signal].segs[iseg]):
            samples -= file.baseline[signal].get_sampled_nodes(first, last)[::skip]
        g = file.segments[iseg, 'Group'] if (group is None) else group
    else:
        samples = numpy.zeros(shape=(last-first+1,), dtype='float32')
    return SourceChunk(file, signal, first, last, included=included, latency=latency,
                       samples=samples, sampling=skip*file.sampling, start=start, baseline_nodes=baseline_nodes, group=group)

def get_file_samples_filtered(file, sig, first, last, included, latency, filter_Hz, skip=1, get_samples=get_file_samples,
                              start=-1, baseline_nodes=True, group=None):
    for f,l,n in file.segmentation.segments:
        if f <= first <= l:
            seg_first, seg_last = f, l
            break
    else:
        print "Can't find segment index?"
        return get_samples(file, sig, first, last, included, latency, skip, baseline_nodes=baseline_nodes, group=group)
    excess = int(round((2 / filter_Hz) / file.sampling)) # 2 wavelengths ok? on either side
    first_got = max(seg_first, first-excess)
    last_got = min(seg_last, last+excess)
    raw_chunk = get_samples(file, sig, first_got, last_got, included, latency, skip, baseline_nodes=baseline_nodes, group=group)
    missing_begin = int(round((excess - (first-first_got)) * 1.0 / skip))
    missing_end = int(round((excess - (last_got-last)) * 1.0 / skip))
    to_skip = int(round(excess * 1.0 / skip))
    datapad = numpy.zeros(shape=(missing_begin + len(raw_chunk.samples) + missing_end,), dtype=numpy.float32)
    datapad[missing_begin : missing_begin + len(raw_chunk.samples)] = raw_chunk.samples
    datapad[:missing_begin] = datapad[missing_begin]
    if missing_end:
        datapad[-missing_end:] = datapad[-missing_end-1]
    filterpad = qubx.fast.filter.filter(datapad, raw_chunk.sampling, filter_Hz)
    if to_skip:
        samples_out = numpy.array(filterpad[to_skip : - to_skip], copy=True)
    else:
        samples_out = filterpad
    return SourceChunk(file=raw_chunk.file, signal=raw_chunk.signal, f=first, l=last, latency=latency,
                       included=raw_chunk.included, samples=samples_out, sampling=raw_chunk.sampling,
                       start=start, baseline_nodes=baseline_nodes, group=group)


def generate_chunk_samples(chunks, main_hold=None, signal=None, maxlen=(1<<16), get_excluded=False,
                           downsample_w=None, downsample_Nsigseg=1, get_idl=False, get_idl_durs=False):
    """Yields (numpy.array(float32) of sampled data, sampling) for each chunk.

    @param chunks: list of L{SourceChunk}
    @param main_hold: L{qubx.task.GTK_Main_Hold} if in a worker thread, or None in the main thread
    @param signal: signal index, or None to use the chunk.signal
    @param maxlen: chunks longer than this will be yielded in pieces; can be None: unlimited
    @param get_excluded: if True, gets data for chunks with .included==False
    @param downsample_w: width of screen in pixels, if you want to maybe skip some for speed
    @param downsample_Nsigseg: total number of segments*signals being processed, to help decide how many points to skip
    @param get_idl: if True, sets chunk.firsts, chunk.lasts, chunk.classes
    @param get_idl_durs: if True and get_idl, sets chunk.durations
"""
    maxchunk = maxlen or max(chunk.n for chunk in chunks)
    if downsample_w is None:
        skip = 1
    else:
        Nsamp = sum(chunk.n for chunk in chunks if get_excluded or chunk.included)
        actual_spp = Nsamp / max(downsample_w, 1)
        skip = max(1, int(round(actual_spp / DOWNSAMPLES_PER_PIX(downsample_Nsigseg, Nsamp))))
    def get_samples_nohold(file, sig, first, last, included, latency, skip, baseline_nodes, group):
        return get_file_samples(file, sig, first, last, included, latency, skip, baseline_nodes=baseline_nodes, group=group)
    def get_samples_hold(file, sig, first, last, included, latency, skip, baseline_nodes, group):
        with main_hold:
            chunk_out = get_samples_nohold(file, sig, first, last, included, latency, skip, baseline_nodes, group)
        return chunk_out
    get_samples = main_hold and get_samples_hold or get_samples_nohold
    for chunk in chunks:
        if chunk.included or get_excluded:
            sig = signal
            if sig is None:
                sig = chunk.signal
            for i in xrange(chunk.f, chunk.l+1, maxchunk):
                if chunk.filter_Hz:
                    chunk_out = get_file_samples_filtered(chunk.file, sig, i, min(chunk.l, i+maxchunk-1), chunk.included, chunk.latency, chunk.filter_Hz, skip, get_samples,
                                                          baseline_nodes=chunk.baseline_nodes, group=chunk.group)
                else:
                    chunk_out = get_samples(chunk.file, sig, i, min(chunk.l, i+maxchunk-1), chunk.included, chunk.latency, skip, chunk.baseline_nodes, chunk.group)
                if get_idl:
                    idl = chunk_out.get_idealization(chunk.signal, False, True, get_idl_durs)
                    chunk_out.firsts, chunk_out.lasts, chunk_out.classes = idl[:3]
                    if get_idl_durs:
                        chunk_out.durations = idl[3]
                yield chunk_out
        else:
            yield chunk


def BuildSession(base_qsf, segs_of_signal, lists, *tables):
    qsf = base_qsf and base_qsf.clone() or qubx.tree.Node()
    if segs_of_signal and segs_of_signal[0]:
        qsf['Sampling'].data = segs_of_signal[0][0].sampling
        qsf['DataGroup'].data = segs_of_signal[0][0].file.group
        BuildSessionIdl(qsf, segs_of_signal)
        BuildSessionExcl(qsf, segs_of_signal)
    for table in tables:
        if qsf.find(table.label):
            qsf.remove(qsf[table.label])
        qsf.append(table_to_tree(table))
    if lists:
        if qsf.find('Lists'):
            qsf.remove(qsf['Lists'])
        qsf.append(lists.to_classic_tree())
        if qsf.find('Lists_Express'):
            qsf.remove(qsf['Lists_Express'])
        qsf.append(lists.to_tree())
    return qsf

def BuildSessionIdl(qsf, segs_of_signal):
    if qsf.find('Idealization'):
        qsf.remove(qsf['Idealization'])
    idlchan = qubx.tree.NullNode()
    for s, segs in enumerate(segs_of_signal):
        points = 0
        idlchan = qsf['Idealization'].insert('Channel', idlchan)
        idlchan['sampling'].data = segs[0].sampling * 1e3
        qsfseg = qubx.tree.NullNode()
        for seg in segs:
            ff, ll, cc, dd, aa, ss = seg.get_idealization(mark_excluded=False, get_fragments=True,
                                                          get_durations=True, get_amps=True, get_stds=True)
            if (len(ff) == 0) or ((len(ff) == 1) and (cc[0] < 0)):
                points += seg.n
                continue
            if seg.f != points:
                ll -= seg.f - points
                ff -= seg.f - points
            qsfseg = idlchan.insert('Segment', qsfseg)
            qsfseg.data = (points, points+seg.n-1)
            points += seg.n
            qsfseg['DwellCount'].data = len(ff)
            qsfseg['Firsts'].data = ff
            qsfseg['Lasts'].data = ll
            qsfseg['Classes'].data = cc
            qsfseg['Durations'].data = dd
            qsfseg['amp'].data = aa
            qsfseg['sd'].data = ss


def BuildSessionFits(qsf, file, segs_of_signal):
    if qsf.find('FitCurves'):
        qsf.remove(qsf['FitCurves'])
    qsf['StoreFitCurves'].data = 1
    idlchan = qubx.tree.NullNode()
    for s, segs in enumerate(segs_of_signal):
        points = 0
        idlchan = qsf['FitCurves'].insert('Channel', idlchan)
        qsfseg = qubx.tree.NullNode()
        for iseg, seg in enumerate(segs):
            qsfseg = idlchan.insert('Segment', qsfseg)
            qsfseg.data = (points, points+seg.n-1)
            ff, ll, cc = file.fits[s].idl.get_dwells_and_gaps(seg.f, seg.l)
            if (len(ff) == 0) or ((len(ff) == 1) and (cc[0] < 0)):
                points += seg.n
                continue
            points += seg.n
            qsfseg['DwellCount'].data = len(ff)
            qsfseg['Firsts'].data = ff
            qsfseg['Lasts'].data = ll
            qsfseg['Classes'].data = cc
            qsfseg['Mean'].data = file.fits[s].idl.segmeans[iseg]
            qsfseg['Std'].data = file.fits[s].idl.segstds[iseg]

def BuildSessionOverlays(qsf, file, segs_of_signal):
    if qsf.find('Overlays'):
        qsf.remove(qsf['Overlays'])
    chan = qubx.tree.NullNode()
    for s, segs in enumerate(segs_of_signal):
        points = 0
        chan = qsf['Overlays'].insert('Channel', chan);
        if file.overlays[s] and file.overlays[s].sels:
            chan['Sels'].data = file.overlays[s].sels
            qsfseg = qubx.tree.NullNode()
            for iseg, seg in enumerate(segs):
                qsfseg = chan.insert('Segment', qsfseg)
                qsfseg.data = (points, points+seg.n-1)
                ff, ll, cc = file.overlays[s].idl.get_dwells_and_gaps(seg.f, seg.l)
                qsfseg['DwellCount'].data = len(ff)
                if len(ff):
                    qsfseg['Firsts'].data = ff
                    qsfseg['Lasts'].data = ll
                    qsfseg['Classes'].data = cc
                points += seg.n

def BuildSessionExcl(qsf, segs_of_signal):
    if qsf.find('Exclusion'):
        qsf.remove(qsf['Exclusion'])
    excl = qsf['Exclusion']
    qsfseg = qubx.tree.NullNode()
    points = 0
    for seg in segs_of_signal[0]:
        qsfseg = excl.insert('Segment', qsfseg)
        qsfseg.data = (points, points+seg.n-1)
        if any(not chunk.included for chunk in seg.chunks):
            ff = []
            ll = []
            cc = []
            segpts = points
            for chunk in seg.chunks:
                ff.append(segpts)
                segpts += chunk.n
                ll.append(segpts - 1)
                if chunk.included:
                    cc.append(1)
                else:
                    cc.append(0)
            qsfseg['DwellCount'].data = len(cc)
            qsfseg['Classes'].data = cc
            qsfseg['Firsts'].data = ff
            qsfseg['Lasts'].data = ll
        points += seg.n

def BuildSessionBaseline(qsf, baselines):
    if qsf.find('BaselineNodes'):
        qsf.remove(qsf['BaselineNodes'])
    bchan = qubx.tree.NullNode()
    for s, baseline in enumerate(baselines):
        bchan = qsf['BaselineNodes'].insert('Channel', bchan)
        bseg = qubx.tree.NullNode()
        for seg in baseline.segs:
            bseg = bchan.insert('Segment', bseg)
            bseg.data = (seg.first, seg.last)
            node = seg.first_node()
            if node:
                points = bseg['Points'].data
                points.setup(qubx.tree.QTR_TYPE_INT, len(seg), 1)
                values = bseg['Values'].data
                values.setup(qubx.tree.QTR_TYPE_FLOAT, len(seg), 1)
                at = 0
                while node:
                    points[at] = node.point
                    values[at] = node.value
                    node = node.next()
                    at += 1


def BuildSessionTrials(qsf, trials):
    if qsf.find('Trials'):
        qsf.remove(qsf['Trials'])
    qsf_trials = qsf['Trials']
    qsf_ts = qubx.tree.NullNode()
    for ts in trials.trialsets:
        if not ts.in_qsf: continue
        qsf_ts = qsf_trials.insert('Trialset', qsf_ts)
        qsf_ts['trialset_name'].data = ts.trialset_name
        last = qsf_ts.append( qubx.table_tree.table_to_tree(ts) )
        last.name = 'Table'
        for r in xrange(ts.size):
            last = qsf_ts.insert(ts.models[r], last)
        

def Save_QDF(path, base_qsf, segs_of_signal, lists, progressf, *tables, **kw):
    tbl = dict((t.label, t) for t in tables)
    signals = tbl['Signals']
    base_path, ext = os.path.splitext(path)
    qsf = BuildSession(base_qsf, segs_of_signal, lists, *tables)
    
    qdf = qubx.tree.Node('DataFile')
    qdf['Sampling'].data = segs_of_signal[0][0].sampling
    qdf['ADDataSize'].data = qsf['ADDataSize'].data = 4
    qdf['ADChannelCount'].data = qsf['ADChannelCount'].data = signals.size
    qdf['ADDataType'].data = qsf['ADDataType'].data = qubx.tree.QTR_TYPE_INT
    qdf.saveAs(path)
    points = 0
    lo = [1e20] * signals.size
    hi = [1e-20] * signals.size
    segout = qubx.tree.NullNode()
    sigs_of_seg = [ [segs_of_signal[c][i] for c in xrange(signals.size)] for i in xrange(len(segs_of_signal[0]))]
    for s, seg_of_signal in enumerate(sigs_of_seg):
        segout = qdf['Segments'].insert('Segment', segout)
        f = points
        l = f + seg_of_signal[0].n - 1
        points += seg_of_signal[0].n
        segout['StartTime'].data = seg_of_signal[0].start * 1e-3
        segout['Channels'].data.setup(qubx.tree.QTR_TYPE_INT, l-f+1, signals.size)
        segout['Channels'].data.preload = False
        for c, seg in enumerate(seg_of_signal):
            for buf_first in xrange(0, seg.n, SAVEBUFLEN):
                buf_last = min(seg.n, buf_first + SAVEBUFLEN) - 1
                buf = seg.get_samples(buf_first, buf_last).samples
                lo[c] = min(lo[c], numpy.min(buf))
                hi[c] = max(hi[c], numpy.max(buf))
    lo = min((x - signals.get(c, 'Offset')) / signals.get(c, 'Scale') for c,x in enumerate(lo))
    hi = max((x - signals.get(c, 'Offset')) / signals.get(c, 'Scale') for c,x in enumerate(hi))
    scaling = (2**24)/max(abs(lo), abs(hi))
    qdf['Scaling'].data = qsf['Scaling'].data = scaling
    qdf.re_map_data()
    segout = qdf['Segments'].find('Segment')
    sigs_of_seg = [ [segs_of_signal[c][i] for c in xrange(signals.size)] for i in xrange(len(segs_of_signal[0]))]
    prog_seg = 1.0 / len(sigs_of_seg)
    for s, seg_of_signal in enumerate(sigs_of_seg):
        if not progressf(s * prog_seg):
            raise KeyboardInterrupt()
        channels = segout['Channels'].storage.data
        prog_sig = prog_seg / len(seg_of_signal)
        for c, seg in enumerate(seg_of_signal):
            if not progressf(s*prog_seg + c*prog_sig):
                raise KeyboardInterrupt()
            for buf_first in xrange(0, seg.n, SAVEBUFLEN):
                if not progressf(s*prog_seg + (c + buf_first*1.0/seg.n)*prog_sig):
                    raise KeyboardInterrupt()
                buf_last = min(seg.n, buf_first + SAVEBUFLEN) - 1
                buf = seg.get_samples(buf_first, buf_last).samples
                buf -= signals.get(c, 'Offset')
                buf *= scaling / signals.get(c, 'Scale')
                channels[buf_first:buf_last+1,c] = buf
        segout = segout.nextSameName()
    ## qsf Scaling, Offset classic-style:
    dchan = qsf['DataChannels']['Channel']
    for c in xrange(signals.size):
        dchan['Scaling'].data = 1 / signals.get(c, 'Scale')
        dchan['Offset'].data = signals.get(c, 'Offset')
        dchan['Name'].data = signals.get(c, 'Name')
        dchan['Units'].data = signals.get(c, 'Units')
        if dchan.sibling:
            dchan = dchan.sibling
        elif (c+1) < signals.size:
            dchan = dchan.parent.insert('Channel', dchan)
    qdf.appendClone(qsf)
    qdf.save()
    del seg, segout, channels
    del qdf
    
    qsf['FileExt'].data = ext.lower()
    qsf.saveAs(base_path+'.qsf', as_copy=True)
    del qsf
SetWriter('.qdf', 'QuB data files', Save_QDF)


def Save_Float32(path, base_qsf, segs_of_signal, lists, progressf, *tables, **kw):
    tbl = dict((t.label, t) for t in tables)
    signals = tbl['Signals']
    base_path, ext = os.path.splitext(path)
    qsf = BuildSession(base_qsf, segs_of_signal, lists, *tables)
    
    fi = open(path, 'wb')
    sigs_of_seg = [ [segs_of_signal[c][i] for c in xrange(signals.size)] for i in xrange(len(segs_of_signal[0]))]
    for s, seg_of_signal in enumerate(sigs_of_seg):
        if not progressf(s * 1.0 / len(sigs_of_seg)):
            raise KeyboardInterrupt()
        channels = numpy.zeros(shape=(seg_of_signal[0].n, len(seg_of_signal)), dtype='float32')
        for c, seg in enumerate(seg_of_signal):
            for buf_first in xrange(0, seg.n, SAVEBUFLEN):
                buf_last = min(seg.n, buf_first + SAVEBUFLEN) - 1
                buf = seg.get_samples(buf_first, buf_last).samples
                channels[buf_first:buf_last+1,c] = buf
        channels.tofile(fi)
    ## qsf Scaling, Offset classic-style:
    dchan = qsf['DataChannels']['Channel']
    for c in xrange(signals.size):
        dchan['Scaling'].data = 1 / signals.get(c, 'Scale')
        dchan['Offset'].data = signals.get(c, 'Offset')
        if dchan.sibling:
            dchan = dchan.sibling
        elif (c+1) < signals.size:
            dchan = dchan.parent.insert('Channel', dchan)
    
    qsf['SinglePrecision'].data = 1
    qsf['ADDataSize'].data = 4
    qsf['ADChannelCount'].data = len(segs_of_signal)
    qsf['FileExt'].data = ext.lower()
    qsf.saveAs(base_path+'.qsf', as_copy=True)
    del qsf
SetWriter('.float32', 'Raw binary (float32)', Save_Float32)


def Save_Int16(path, base_qsf, segs_of_signal, lists, progressf, *tables, **kw):
    tbl = dict((t.label, t) for t in tables)
    signals = tbl['Signals']
    base_path, ext = os.path.splitext(path)
    qsf = BuildSession(base_qsf, segs_of_signal, lists, *tables)

    fi = open(path, 'wb')
    points = 0
    lo = [1e20] * signals.size
    hi = [1e-20] * signals.size
    sigs_of_seg = [ [segs_of_signal[c][i] for c in xrange(signals.size)] for i in xrange(len(segs_of_signal[0]))]
    for s, seg_of_signal in enumerate(sigs_of_seg):
        f = points
        l = f + seg_of_signal[0].n - 1
        points += seg_of_signal[0].n
        for c, seg in enumerate(seg_of_signal):
            for buf_first in xrange(0, seg.n, SAVEBUFLEN):
                buf_last = min(seg.n, buf_first + SAVEBUFLEN) - 1
                buf = seg.get_samples(buf_first, buf_last).samples
                lo[c] = min(lo[c], numpy.min(buf))
                hi[c] = max(hi[c], numpy.max(buf))
    lo = min((x - signals.get(c, 'Offset')) / signals.get(c, 'Scale') for c,x in enumerate(lo))
    hi = max((x - signals.get(c, 'Offset')) / signals.get(c, 'Scale') for c,x in enumerate(hi))
    scaling = 32767/max(abs(lo), abs(hi))
    qsf['Scaling'].data = scaling
    sigs_of_seg = [ [segs_of_signal[c][i] for c in xrange(signals.size)] for i in xrange(len(segs_of_signal[0]))]
    for s, seg_of_signal in enumerate(sigs_of_seg):
        if not progressf(s * 1.0 / len(sigs_of_seg)):
            raise KeyboardInterrupt()
        channels = numpy.zeros(shape=(seg_of_signal[0].n, len(seg_of_signal)), dtype='int16')
        for c, seg in enumerate(seg_of_signal):
            for buf_first in xrange(0, seg.n, SAVEBUFLEN):
                buf_last = min(seg.n, buf_first + SAVEBUFLEN) - 1
                buf = seg.get_samples(buf_first, buf_last).samples
                buf -= signals.get(c, 'Offset')
                buf *= scaling / signals.get(c, 'Scale')
                channels[buf_first:buf_last+1,c] = buf
        channels.tofile(fi)
    ## qsf Scaling, Offset classic-style:
    dchan = qsf['DataChannels']['Channel']
    for c in xrange(signals.size):
        dchan['Scaling'].data = 1 / signals.get(c, 'Scale')
        dchan['Offset'].data = signals.get(c, 'Offset')
        if dchan.sibling:
            dchan = dchan.sibling
        elif (c+1) < signals.size:
            dchan = dchan.parent.insert('Channel', dchan)
    
    qsf['SinglePrecision'].data = 0
    qsf['ADDataSize'].data = 2
    qsf['ADChannelCount'].data = len(segs_of_signal)
    qsf['FileExt'].data = ext.lower()
    qsf.saveAs(base_path+'.qsf', as_copy=True)
    del qsf
SetWriter('.int16', 'Raw binary (int16)', Save_Int16)


def Save_TXT_Base(separator):
    def Save(path, base_qsf, segs_of_signal, lists, progressf, *tables, **kw):
        tbl = dict((t.label, t) for t in tables)
        signals = tbl['Signals']
        base_path, ext = os.path.splitext(path)
        qsf = BuildSession(base_qsf, segs_of_signal, lists, *tables)
        
        total_lines = written_lines = 0
        for seg in segs_of_signal[0]:
            total_lines += seg.n
        
        file = open(path, 'w')
        file.write('Time%s' % separator)
        file.write(separator.join(signals.get(c, 'Name') for c in xrange(signals.size)))
        file.write('\n')
        sampling = segs_of_signal[0][0].sampling
        sigs_of_seg = [ [segs_of_signal[c][i] for c in xrange(signals.size)] for i in xrange(len(segs_of_signal[0]))]
        for s, seg_of_signal in enumerate(sigs_of_seg):
            if s:
                file.write('\n')
            for buf_first in xrange(0, seg_of_signal[0].n, SAVEBUFLEN):
                buf_last = min(seg_of_signal[0].n, buf_first + SAVEBUFLEN) - 1
                time = numpy.arange((buf_last - buf_first + 1), dtype='float32')
                time += buf_first
                time *= sampling
                time += 1e-3*(seg_of_signal[0].start - seg_of_signal[0].file.segments[0, 'Start'])
                columns = [time]
                columns.extend([seg.get_samples(buf_first, buf_last).samples for seg in seg_of_signal])
                for tup in itertools.izip(*columns):
                    file.write(separator.join("%.9g" % x for x in tup))
                    file.write('\n')
                    written_lines += 1
                if not progressf(written_lines * 1.0 / total_lines):
                    raise KeyboardInterrupt()
        qsf['TimeColumn'].data = int(True)
        qsf['Sampling'].data = sampling
        qsf['FileExt'].data = ext.lower()
        qsf.saveAs(base_path+'.qsf', as_copy=True)
        del qsf
    return Save
Save_TXT = Save_TXT_Base('\t')
Save_CSV = Save_TXT_Base(',')
SetWriter('.txt', 'Text files', Save_TXT)
SetWriter('.csv', 'Comma-separated ext files', Save_CSV)


def ReadSessionIdl(data, sess, addsegs=False, progressf=lambda pct:True):
    """
Fills in data.ideal (and optionally data.segmentation) from sess, a
U{QSF<http://www.qub.buffalo.edu/qubdoc/files/qsf.html>} tree.
"""
    for i, chan in enumerate(qubx.tree.children(sess['Idealization'], 'Channel')):
        for seg in qubx.tree.children(chan, 'Segment'):
            if not progressf(0.999): return
            if addsegs and (i == 0):
                data.segmentation.add_seg(seg.data[0], seg.data[1], seg.data[0]*data.sampling*1e3)
            dwell_count = 0
            if seg.find('DwellCount') and seg['DwellCount'].data:
                dwell_count = seg['DwellCount'].data[0]
            elif seg.find('Firsts') and seg['Firsts'].data:
                dwell_count = len(seg['Firsts'].data)
            if dwell_count:
                data.ideal[i].idl.set_dwells(dwell_count,
                                             seg['Firsts'].storage.data,
                                             seg['Lasts'].storage.data,
                                             seg['Classes'].storage.data)
                j = data.segmentation.index_at(seg.data[0])
                data.ideal[i].seg[j].amp = seg['amp'].data[:]
                data.ideal[i].seg[j].std = seg['sd'].data[:]
    sess.remove(sess['Idealization']) # save memory

def ReadSessionFits(data, sess, progressf=lambda pct:True):
    data.store_fit_curves = sess.find('StoreFitCurves').data and sess['StoreFitCurves'].data[0] or 0
    for i, chan in enumerate(qubx.tree.children(sess.find('FitCurves'), 'Channel')):
        for j, seg in enumerate(qubx.tree.children(chan, 'Segment')):
            if not progressf(0.999): return
            dwell_count = 0
            if seg.find('DwellCount') and seg['DwellCount'].data:
                dwell_count = seg['DwellCount'].data[0]
            elif seg.find('Firsts') and seg['Firsts'].data:
                dwell_count = len(seg['Firsts'].data)
            if dwell_count:
                data.fits[i].idl.set_dwells(dwell_count,
                                             seg['Firsts'].storage.data,
                                             seg['Lasts'].storage.data,
                                             seg['Classes'].storage.data)
                data.fits[i].idl.segmeans[j] = seg['Mean'].storage.data.flatten()
                data.fits[i].idl.segstds[j] = seg['Std'].storage.data.flatten()
    sess.remove(sess['FitCurves']) # save memory

def ReadSessionOverlays(data, sess, progressf=lambda pct:True):
    for i, chan in enumerate(qubx.tree.children(sess.find('Overlays'), 'Channel')):
        data_sel = chan.find('Sels').data
        if data_sel:
            overlays = data.get_overlays(i)
            for r in xrange(data_sel.rows):
                overlays.set_source(* data_sel.row(r)[:2])
            for j, seg in enumerate(qubx.tree.children(chan, 'Segment')):
                if not progressf(0.999): return
                dwell_count = 0
                if seg.find('DwellCount') and seg['DwellCount'].data:
                    dwell_count = seg['DwellCount'].data[0]
                elif seg.find('Firsts') and seg['Firsts'].data:
                    dwell_count = len(seg['Firsts'].data)
                if dwell_count:
                    overlays.idl.set_dwells(dwell_count,
                                            seg['Firsts'].storage.data,
                                            seg['Lasts'].storage.data,
                                            seg['Classes'].storage.data)
    sess.remove(sess['Overlays']) # save memory

def ReadSessionExcl(data, sess, progressf=lambda pct:True):
    for j,seg in enumerate(qubx.tree.children(sess.find('Exclusion'), 'Segment')):
        if not progressf(0.999): return
        dwell_count = 0
        if seg.find('DwellCount') and seg['DwellCount'].data:
            dwell_count = seg['DwellCount'].data[0]
        elif seg.find('Firsts') and seg['Firsts'].data:
            dwell_count = len(seg['Firsts'].data)
        if dwell_count:
            data.exclusion.idl.set_dwells(dwell_count,
                                          seg['Firsts'].storage.data,
                                          seg['Lasts'].storage.data,
                                          seg['Classes'].storage.data)
    data.exclusion.OnChange(data.exclusion)

def ReadSessionBaseline(data, sess, progressf=lambda pct:True):
    if sess.find('BaselineNodes'):
        for i, chan in enumerate(qubx.tree.children(sess['BaselineNodes'], 'Channel')):
            for j, seg in enumerate(qubx.tree.children(chan, 'Segment')):
                if not progressf(0.999): return
                points, values = seg.find('Points'), seg.find('Values')
                if points.data and values.data:
                    data.baseline[i].add_nodes(points.storage.data.T[0,:], values.storage.data.T[0,:])
    for i, basechan in enumerate(data.baseline):
        if any(len(bseg) for bseg in basechan.segs):
            break # skip the for/else clause
    else: # no baseline so far, try QuB Classic (values as integers)
        for i, chan in enumerate(qubx.tree.children(sess.find('Baseline'), 'BaselineSegments')):
            for j, seg in enumerate(qubx.tree.children(chan, 'BaselineSegment')):
                if not progressf(0.999): return
                points, values = seg.find('Points'), seg.find('Values')
                if points.data and values.data:
                    pp, vv = points.storage.data.T[0,:], values.storage.data.T[0,:].astype('float32')
                    vv *= - data.signals[i, 'Scale'] / sess['Scaling'].data[0]
                    vv += data.signals[i, 'Offset']
                    data.baseline[i].add_nodes(pp, vv)
    sess.remove(sess['BaselineNodes']) # save memory

def ReadSessionTrials(data, sess, progressf=lambda pct:True):
    for qsf_ts in qubx.tree.children(sess.find('Trials'), 'Trialset'):
        ts_name = str(qsf_ts.find('trialset_name').data)
        if ts_name and qsf_ts.find('Table'):
            ts = data.trials.get_trialset(ts_name, True)
            qubx.table_tree.tree_to_table(qsf_ts['Table'], ts)
            for r, model in enumerate(qubx.tree.children(qsf_ts, 'ModelFile')):
                ts.models[r] = model
    sess.remove(sess['Trials']) # save memory


def read_signal_overlay(samples, data, overlay, analog, sel_index):
    cached = overlay.data[sel_index]
    if cached is None:
        cached = overlay.data[sel_index] = analog.read(*overlay.sels[sel_index])
    Nover = len(cached)
    Nsample = len(samples)
    for at in xrange(0, Nsample, Nover):
        batch = min(Nsample-at, Nover)
        samples[at:at+batch] = cached[:batch]


class QubData_Analog(object):
    """Base class for one analog signal."""
    def read(self, first, last, latency=0, skip=1):
        """Override this method to return a numpy.array(float32) of data samples from within one segment.

@param latency: the data is shifted to the right by this many samples; repeats seg. first/last sample as needed
@param skip: e.g. 1: every point, 2: every other point, 3: every third point, ...
"""
        pass
    def read_overlaid(self, datafile, signal, first, last, latency=0, skip=1):
        """Returns samples, like L{QubData_Analog.read}, but replaces regions marked in datafile.overlays[signal] with data from elsewhere in the file."""
        overlays = datafile.overlays[signal]
        if overlays:
            ff, ll, cc = overlays.idl.get_dwells_and_gaps(first, last)
            Nd = len(cc)
            if Nd:
                samples = numpy.zeros(shape=(last-first+1,), dtype='float32')
                for f, l, c in izip(ff, ll, cc):
                    if c < 0:
                        samples[f-first:l+1-first] = self.read(f, l, latency, skip)
                    else:
                        read_signal_overlay(samples[f-first:l+1-first], datafile, overlays, self, c)
                return samples
        return self.read(first, last, latency, skip)


def parse_computed_signal(expr, source_names):
    try:
        obj = qubx.pyenv.env.eval_str(expr, False)
        if hasattr(obj, 'shape') or hasattr(obj, 'get_samples'):
            return acceptF(static=[], custom=False, typ=None)(expr)
    except:
        pass # not a simple array or proto signal
    func = acceptF(static=[], custom=True)(expr)
    allowed = set(['x', 't'] + list(source_names))
    for name in func.args:
        if not (name in allowed):
            raise NameError('No signal named "%s" in computed signal: %s' % (name, func.expr))
    return func

class QubData_Analog_Computed(QubData_Analog):
    def __init__(self, func, sources, data, index):
        self.func = func
        self.sources = sources
        self.data = data
        self.index = index
    def read(self, first, last, latency=0, skip=1):
        if not self.func.args: # names a numpy array congruent to entire file, or sim proto signal
            arr = numpy.zeros(shape=(last-first+1,), dtype='float32')
            src = self.func()
            if hasattr(src, 'get_samples'):
                src.get_samples(self.sampling, first, last, arr)
            else:
                if len(src) > first:
                    src = src[first:min(len(src), last+1)]
                    arr[:len(src)] = src
            if skip > 1:
                arr = arr[::skip]
            return arr
        source_samples = []
        for arg in self.func.args:
            if arg in ('x', 't'):
                arr = numpy.arange(0, (last-first+1), skip, dtype='float32')
                arr += first
                arr *= self.sampling
                source_samples.append(arr)
            elif arg in self.sources:
                source_samples.append(self.sources[arg].read_overlaid(self.data, self.index, first, last, latency, skip))
            else:
                raise NameError('"%s" is unkown in computed signal expression: %s' % (arg, func.expr))
        try:
            return self.func(*source_samples)
        except: # the function doesn't broadcast to numpy arrays?
            # return numpy.array([self.func(*args) for args in izip(*source_samples)], dtype=numpy.float32)
            arr = numpy.zeros(shape=source_samples[0].shape, dtype=numpy.float32)
            for i, args in enumerate(izip(*source_samples)):
                try:
                    arr[i] = self.func(*args)
                except:
                    pass
            return arr


def read_with_latency(read, first, last, latency, skip, iseg, segfirst, seglast):
    lat_f, lat_l = first - latency, last - latency
    lat_fill_pre = max(0, segfirst - lat_f)
    lat_fill_post = max(0, lat_l - seglast)
    lat_f += lat_fill_pre
    lat_l -= lat_fill_post
    samples = read(iseg, lat_f-segfirst, lat_l-segfirst, skip)
    if lat_fill_pre or lat_fill_post:
        samples, sub = numpy.zeros(shape=(last-first+1,), dtype=numpy.float32), samples
        if lat_fill_pre:
            samples[:lat_fill_pre] = sub[0]
        if lat_fill_post:
            samples[-lat_fill_post:] = sub[-1]
        samples[lat_fill_pre:-lat_fill_post or len(samples)] = sub[:]
    return samples


class QubData_QDF_Analog(QubData_Analog):
    def __init__(self, data, channel):
        """
        @param data: L{QubData_QDF}
        @param channel: integer 0-based signal index
"""
        self.data = data
        self.channel = channel
    def read(self, first, last, latency=0, skip=1):
        ifirst, ilast = [self.data.segmentation.index_at(x) for x in (first, last)]
        if ifirst != ilast:
            raise Exception("QDF: can't read() from multiple segments at once")
        segfirst = self.data.segmentation.segments[ifirst][0]
        seglast = self.data.segmentation.segments[ifirst][1]
        return read_with_latency(self.read_in_seg, first, last, latency, skip, ifirst, segfirst, seglast)
    def read_in_seg(self, iseg, first, last, skip):
        channels = self.data.seg[iseg].channels
        if channels.storage.data is None:
            samples = channels.data.read_rows(first, last)[::skip, self.channel].flatten().astype('float32')
        elif (not (channels.storage.data is None)) and len(channels.storage.data):
            samples = channels.storage.data[first:last+1:skip, self.channel].flatten().astype('float32')
        else:
            if qubx.global_namespace.DEBUG:
                print 'missing segment %i data of %s' % (iseg, self.data.path)
            samples = numpy.zeros(shape=((last-first+1)/skip, 1), dtype='float32')
        samples /= self.data.scaling
        scale = self.data.signals.get(self.channel, 'Scale')
        if scale != 1.0:
            samples *= scale
        offset = self.data.signals.get(self.channel, 'Offset')
        if offset != 0.0:
            samples += offset
        return samples


class QubData_QDF(QubData):
    """One open U{QDF<http://www.qub.buffalo.edu/qubdoc/files/qdf.html>}
(or U{QSF<http://www.qub.buffalo.edu/qubdoc/files/qsf.html>}) data file.

@ivar qdf: L{qubx.tree.Node_numpy} the actual file
@ivar scaling: float_val = int_val / scaling
"""
    def __init__(self, path, progressf, sess):
        QubData.__init__(self)
        self.path = path
        base, ext = os.path.splitext(path)
        if ext.lower() == '.qsf':
            self.qdf = qubx.tree.NullNode()
        else:
            self.qdf = qubx.tree.Open(path, True)

        if sess.child.isNull:
            sess = qubx.tree.AS_FLAVOR(self.qdf.find('SessionFile'), 'numpy', True)
        self.qsf = sess
        
        self.sampling = ((sess.find('Sampling') and sess['Sampling'].data[0])
                         or (self.qdf.find('Sampling') and self.qdf['Sampling'].data[0])
                         or 1.0)
        self.scaling = ((sess.find('Scaling') and sess['Scaling'].data[0])
                        or (self.qdf.find('Scaling') and self.qdf['Scaling'].data[0])
                        or 1.0)
        self.nchan = ((sess.find('ADChannelCount') and sess['ADChannelCount'].data[0])
                      or (self.qdf.find('ADChannelCount') and self.qdf['ADChannelCount'].data[0])
                      or 0)
        c = 0
        for chan in qubx.tree.children(sess.find('DataChannels'), 'Channel'):
            self.signals.append({'Name' : str(chan['Name'].data)})
#                                 'Units' : str(chan['Units'].data),
#                                 'Scale' : 1.0 / (chan['Scaling'].data and chan['Scaling'].data[0] or 1.0),
#                                 'Offset' : chan['Offset'].data and chan['Offset'].data[0] or 0.0,
#                                 'Filter' : bool(filters and filters[c] or False),
#                                 'Filter Freq' : freqs and freqs[c] or 1.0})
            c += 1
        while c < self.nchan:
            self.signals.append({'Name' : 'Channel %i' % (c+1)})
#                                 'Scale' : 1.0,
#                                 'Offset' : 0.0})
            c += 1
        
        if self.qdf:
            last = -1
            i = 0
            for seg in qubx.tree.children(self.qdf.find('Segments'), 'Segment'):
                channels = seg['Channels']
                if not channels.data:
                    continue
                first = last + 1
                start = seg.find('StartTime').data and (1e3 * seg['StartTime'].data[0]) or (first*1e3*self.sampling)
                last = last + (channels.data and channels.data.rows or (first - 1))
                if first <= last:
                    self.segmentation.add_seg(first, last, start)
                    self.seg[i].channels = channels
                    i += 1
        
            for i in xrange(self.nchan):
                self.set_analog(i, QubData_QDF_Analog(self, i))
            
        self.read_session(sess, progressf)
    def dispose(self):
        QubData.dispose(self)
        # lose refs to sub-trees before the root, otherwise subtrees are copied to ram
        del self.seg
        del self.qdf
        

def OpenQDF(path, progressf):
    base, ext = os.path.splitext(path)
    sess = qubx.tree.Open(base+'.qsf', True)
    sess.close() # release file, keep in mem
    return QubData_QDF(path, progressf, sess)
SetReader('.qdf', 'QuB data files', OpenQDF)

def OpenQSF(path, progressf):
    base, ext = os.path.splitext(path)
    sess = qubx.tree.Open(path, True)
    sess.close() # release file, keep in mem
    orig_path = base+str(sess['FileExt'].data)
    if os.path.exists(orig_path):
        try:
            return Open(orig_path, progressf) # open that file if we can
        except:
            pass                              # or just use the bare session here
    return QubData_QDF(path, progressf, sess)
SetReader('.qsf', 'QuB session files', OpenQSF)


class QubData_Bin_Analog(QubData_Analog):
    def __init__(self, data, channel):
        """
        @param data: L{QubData_Bin}
        @param channel: integer 0-based signal index
"""
        self.data = data
        self.channel = channel
    def read(self, first, last, latency=0, skip=1):
        ifirst, ilast = [self.data.segmentation.index_at(x) for x in (first, last)]
        if ifirst != ilast:
            raise Exception("can't read() from multiple segments at once") # should be just one seg anyway
        segfirst = self.data.segmentation.segments[ifirst][0]
        seglast = self.data.segmentation.segments[ifirst][1]
        return read_with_latency(self.read_in_seg, first, last, latency, skip, ifirst, segfirst, seglast)
    def read_in_seg(self, iseg, first, last, skip):
        samples = self.data.buf[first:last+1:skip, self.channel].flatten().astype('float32')
        if self.data.scaling != 1.0:
            samples /= self.data.scaling
        scale = self.data.signals.get(self.channel, 'Scale')
        if scale != 1.0:
            samples *= scale
        offset = self.data.signals.get(self.channel, 'Offset')
        if offset != 0.0:
            samples += offset
        return samples
            

class QubData_Bin(QubData):
    """One open raw binary data file.

@ivar scaling: float_val = int_val / scaling
"""
    def __init__(self, path, progressf):
        QubData.__init__(self)
        self.path = path
        self.fileno = self.mmap = None
        exists = os.path.exists(path)
        size = 0
        if exists:
            #st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size, st_atime, st_mtime, st_ctime = os.stat(path)
            size = os.stat(path)[6]
            if size:
                self.fileno = os.open(path, os.O_RDONLY)
                self.mmap = mmap.mmap(self.fileno, size, access=mmap.ACCESS_READ)
        if not self.mmap:
            raise Exception('failed to open %s'%path)
        base, ext = os.path.splitext(path)
        sess = qubx.tree.Open(base+'.qsf', True)
        sess.close() # release file, keep in mem
        self.qsf = sess
        self.sampling = ((sess.find('Sampling') and sess['Sampling'].data[0])
                         or 1.0e-4)
        self.scaling = ((sess.find('Scaling') and sess['Scaling'].data[0])
                        or 1.0)
        self.floating = ((sess.find('SinglePrecision').data and bool(sess['SinglePrecision'].data[0]))
                         or False)
        self.bytes = ((sess.find('ADDataSize') and sess['ADDataSize'].data[0])
                      or (self.floating and 4 or 2))
        self.nchan = ((sess.find('ADChannelCount') and sess['ADChannelCount'].data[0])
                      or 1)
        self.sampling, self.scaling, self.floating, self.bytes, self.nchan = \
            RequestBinaryParams(self.sampling, self.scaling, self.floating, self.bytes, self.nchan)
        if self.floating:
            self.scaling = 1.0
        sess['Sampling'].data = self.sampling
        sess['Scaling'].data = self.scaling
        sess['SinglePrecision'].data = int(self.floating)
        sess['ADDataSize'].data = self.bytes
        sess['ADChannelCount'].data = self.nchan
        c = 0
        for chan in qubx.tree.children(sess.find('DataChannels'), 'Channel'):
            self.signals.append({'Name' : str(chan['Name'].data)})
#                                 'Units' : str(chan['Units'].data),
#                                 'Scale' : 1.0 / (chan['Scaling'].data and chan['Scaling'].data[0] or 1.0),
#                                 'Offset' : chan['Offset'].data and chan['Offset'].data[0] or 0.0,
#                                 'Filter' : bool(filters and filters[c] or False),
#                                 'Filter Freq' : freqs and freqs[c] or 1.0})
            c += 1
        while c < self.nchan:
            self.signals.append({'Name' : 'Channel %i' % (c+1)})
#                                 'Scale' : 1.0,
#                                 'Offset' : 0.0})
            c += 1
        
        nsample = int(size / self.bytes)
        nframe = nsample / self.nchan
        self.segmentation.add_seg(0, nframe-1, 0.0)
        
        if self.floating:
            if self.bytes == 8:
                dtype = 'float64'
            elif self.bytes == 4:
                dtype = 'float32'
            else:
                raise Exception('There are no 2-byte floats')
        else:
            dtype = 'int%d'%(8*self.bytes)
        
        self.buf = numpy.frombuffer(self.mmap, dtype, nsample, 0)
        self.buf = self.buf.reshape((nframe, self.nchan))
        
        for i in xrange(self.nchan):
            self.set_analog(i, QubData_Bin_Analog(self, i))
            
        self.read_session(sess, progressf)
        

class QubData_LDT_Analog(QubData_Analog):
    def __init__(self, data):
        """
        @param data: L{QubData_LDT}
"""
        self.data = data
    def read(self, first, last, latency=0, skip=1):
        ifirst, ilast = [self.data.segmentation.index_at(x) for x in (first, last)]
        if ifirst != ilast:
            raise Exception("can't read() from multiple segments at once") # should be just one seg anyway
        segfirst = self.data.segmentation.segments[ifirst][0]
        seglast = self.data.segmentation.segments[ifirst][1]
        return read_with_latency(self.read_in_seg, first, last, latency, skip, ifirst, segfirst, seglast)
    def read_in_seg(self, iseg, first, last, skip):
        samples = self.data.segbuf[iseg][first:last+1:skip].astype('float32')
        if self.data.scaling != 1.0:
            samples /= self.data.scaling
        scale = self.data.signals.get(0, 'Scale')
        if scale != 1.0:
            samples *= scale
        offset = self.data.signals.get(0, 'Offset')
        if offset != 0.0:
            samples += offset
        return samples


class QubData_LDT(QubData):
    """One open LDT file (classic QuB data)

@ivar scaling: float_val = int_val / scaling
"""
    def __init__(self, path, progressf):
        QubData.__init__(self)
        self.path = path
        self.fileno = self.mmap = None
        exists = os.path.exists(path)
        size = 0
        if exists:
            #st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size, st_atime, st_mtime, st_ctime = os.stat(path)
            size = os.stat(path)[6]
            if size:
                self.fileno = os.open(path, os.O_RDONLY)
                self.mmap = mmap.mmap(self.fileno, size, access=mmap.ACCESS_READ)
        if not self.mmap:
            raise Exception('failed to open %s'%path)
        
        self.signals.append({'Name' : 'Current', 'Units' : 'pA'})
        
        read_val = lambda offset, typ: numpy.frombuffer(self.mmap, typ, 1, offset)[0]
        segpad = read_val(0, '<i4')
        self.sampling = 1e-6 * read_val(4, '<u2')
        self.scaling = 1e-2 * read_val(6, '<u2')
        self.segbuf = []
        offset = 8
        points = 0
        while offset < size:
            start = 1e-3 * read_val(offset, '<i4')
            offset += 4
            nsample = read_val(offset, '<i4') - 2*segpad
            offset += 4
            self.segmentation.add_seg(points, points+nsample-1, start)
            self.segbuf.append(numpy.frombuffer(self.mmap, '<i2', nsample, offset+2*segpad))
            points += nsample
            offset += 2*(nsample+2*segpad)
        
        self.set_analog(0, QubData_LDT_Analog(self))
            
        base, ext = os.path.splitext(path)
        sess = qubx.tree.Open(base+'.qsf', True)
        sess.close() # release file, keep in mem
        self.qsf = sess
        
        self.read_session(sess, progressf)

    def __del__(self):
        if self.mmap:
            self.mmap.close()
            self.mmap = None
        if self.fileno:
            os.close(self.fileno)

SetReader('.ldt', 'QuB classic data files', QubData_LDT)
        

class QubData_Scratch_Analog(QubData_Analog):
    """Reads data from a qub express scratch file (simulation)."""
    def __init__(self, data, channel):
        self.data = data
        self.channel = channel
        self.segments = [] # appended by QubData_Scratch, matching data.segmentation
    def read(self, first, last, latency=0, skip=1):
        ifirst, ilast = [self.data.segmentation.index_at(x) for x in (first, last)]
        if ifirst != ilast:
            raise Exception("scratch: can't read() from multiple segments at once")
        try:
            segfirst = self.data.segmentation.segments[ifirst][0]
            seglast = self.data.segmentation.segments[ifirst][1]
            samples = read_with_latency(self.read_in_seg, first, last, latency, skip, ifirst, segfirst, seglast)
        except:
            #traceback.print_exc()
            samples = numpy.array(numpy.random.randn((last-first+1)/skip), copy=True, dtype='float32')
            samples *= 0.15
        return samples
    def read_in_seg(self, iseg, first, last, skip):
        samples = numpy.array(self.segments[iseg][first:last+1:skip].flatten(), dtype='float32', copy=True)
        scale = self.data.signals.get(self.channel, 'Scale')
        if scale != 1.0:
            samples *= scale
        offset = self.data.signals.get(self.channel, 'Offset')
        if offset != 0.0:
            samples += offset
        return samples


# TODO: scope respond to Signal: Scale

class QubData_Blank(QubData):
    def __init__(self, label):
        QubData.__init__(self)
        self.path = label
        self.signals = self.segments = self.constants = self.stimuli = None


class QubData_Scratch(QubData):
    """One open qub express scratch file (simulation); .tree can be repeatedly re-assigned.

@ivar tree: a L{qubx.tree.Node_numpy} in scratch format, or L{qubx.tree.NullNode_numpy}()"""
    def __init__(self, label):
        """@param label: a description to show in place of the file name."""
        QubData.__init__(self)
        self.path = label
        if (not label) or (label[0] == '<'):
            self.group = 0 # keep simulations out of the file list
        self.temppath = None
        self.sampling = 1e-4
        self.signals.append({'Name' : 'Current',
                             'Units' : 'pA',
                             'Scale' : 1.0,
                             'Offset' : 0.0})
        self.set_analog(0, QubData_Scratch_Analog(self, 0))
        self.segmentation.add_seg(0, 99999, 0.0)
        self.__tree = None
    def set_tree(self, x):
        if x:
            self.undoIdl.clear()
            nchan = x['SignalCount'].data[0]
            while nchan < self.signals.size:
                self.signals.remove(self.signals.size-1)
            # Signals
            sig = x['Signals'].find('Signal')
            for c in xrange(nchan):
                if sig:
                    name, units = str(sig['Name'].data), str(sig['Units'].data)
                else:
                    name, units = ('Signal %i'%(c+1), '')
                sig = sig.next('Signal')
                if c == self.signals.size:
                    self.signals.append({'Name' : name, 'Units' : units})
                else:
                    self.signals.set(c, 'Name', name)
                    self.signals.set(c, 'Units', units)
                    self.ideal[c].idl.clear()
                    self.fits[c].idl.clear()
                self.set_analog(c, QubData_Scratch_Analog(self, c))
            self.sampling = x['Sampling'].data[0]
            # Segments
            segm = []
            first = 0
            for i, seg in enumerate(qubx.tree.children(x['Segments'], 'Segment')):
                sample_count = seg['SampleCount'].data[0]
                last = first + sample_count - 1
                segm.append((first, last, sample_count))
                first = last + 1
                for c, sig in enumerate(qubx.tree.children(seg['Signals'], 'Signal')):
                    self.analog[c].segments.append( sig['Samples'].storage.data )
            if tuple(segm) != tuple(self.segmentation.segments):
                self.segmentation.clear()
                for i, fln in enumerate(segm):
                    self.segmentation.add_seg(fln[0], fln[1], fln[0]*self.sampling*1e3)
            # Ideal
            for s, sig in enumerate(qubx.tree.children(x['IdealSignals'], 'IdealSignal')):
                if sig['SignalIndex'].data:
                    c = sig['SignalIndex'].data[0]
                else:
                    c = s
                for j, seg in enumerate(qubx.tree.children(sig['Segments'], 'Segment')):
                    self.ideal[c].idl.set_dwells(seg['DwellCount'].data[0],
                                                   seg['Firsts'].storage.data,
                                                   seg['Lasts'].storage.data,
                                                   seg['Classes'].storage.data)
                    self.ideal[c].seg[j].amp = seg['amp'].data[:]
                    self.ideal[c].seg[j].std = seg['sd'].data[:]
            # Constants
            for con in qubx.tree.children(x['Constants']):
                try:
                    self.constants.set( self.constants.index(con.name), 'Value', con.data[0] )
                except:
                    self.constants.append({'Name' : con.name,
                                           'Value': con.data[0]})
        else:
            self.segmentation.clear()
        self.__tree = x
        if self.temppath:
            try:
                print 'del',self.temppath,'...'
                os.remove(self.temppath)
                print '   ok'
            except:
                pass
            self.temppath = None
        self.OnChangeSamples(self, -1)
        self.OnChangeIdealization(self, -1)
        self.OnChangeFits(self, -1)
        setup_segments_list(self)
    tree = property(lambda self: self.__tree, lambda self, x: self.set_tree(x))




class TextStats(object):
    def __init__(self, tokenize, convert, debug=False):
        self.tokenize = tokenize
        self.convert = convert
        self.debug = debug
        self.min_fields = None
        self.max_fields = 0
        self.numeric_lines = 0
        self.headers = []
        self.sampling = None
        self.last_time = None
        self.seg_lens = []
        self.seg_lines = 0
    def score(self):
        self.col_count = self.max_fields
        self.signal_count = self.col_count - int(bool(self.sampling))
        return (self.max_fields - (self.min_fields or 0)) + self.numeric_lines + int(bool(self.headers)) + int(bool(self.sampling))
    def finish(self):
        if self.seg_lines:
            self.seg_lens.append(self.seg_lines)
            self.seg_lines = 0
        self.col_count = self.max_fields
        self.signal_count = self.col_count - int(bool(self.sampling))
    def add(self, line):
        fields = self.tokenize(line)
        if len(fields) and fields[0]:
            if (self.min_fields is None) or (self.min_fields > len(fields)):
                self.min_fields = len(fields)
            self.max_fields = max(self.max_fields, len(fields))
            try:
                if self.debug:
                    print fields
                nums = [self.convert(x) for x in [xx.strip() for xx in fields] if x]
                if self.debug:
                    print 'nums:',nums
                self.seg_lines += 1
                self.numeric_lines += 1
                if self.sampling is None:
                    if not (self.last_time is None):
                        self.sampling = nums[0] - self.last_time
                elif self.sampling:
                    if (not (self.last_time is None)) and (abs(self.sampling - (nums[0] - self.last_time)) > (self.sampling*0.1)):
                        self.sampling = 0.0
                self.last_time = nums[0]
            except ValueError:
                if fields and not self.numeric_lines:
                    if self.debug:
                        print 'headers:',fields
                    self.headers = fields
        else:
            self.last_time = None
            if self.seg_lines:
                self.seg_lens.append(self.seg_lines)
                self.seg_lines = 0

def float_decimal_comma(s):
    return float(s.replace(',', '.'))

def OpenTXT(path, progressf):
    #temppath = os.path.join(tempfile.gettempdir(), os.path.split(path)[1])
    #temppath = os.path.splitext(temppath)[0] + '.qubtmp'
    tok_tab = lambda s: s.split('\t')
    re_comma = re.compile(r", *")
    tok_comma = lambda s: re_comma.split(s)
    re_spaces = re.compile(r" +")
    tok_spaces = lambda s: re_spaces.split(s)
    formats = [TextStats(tok_tab, float),   TextStats(tok_tab, float_decimal_comma),
               TextStats(tok_comma, float), TextStats(tok_comma, float_decimal_comma),
               TextStats(tok_spaces, float), TextStats(tok_spaces, float_decimal_comma)]
    lines = [line.strip() for line in open(path, 'r')]
    for line in lines[:min(len(lines), 10)]:
        for format in formats:
            format.add(line)
    if not [format for format in formats if format.numeric_lines]:
        return QubData_Bin(path, progressf)
    scores = numpy.array([format.score() for format in formats])
    format = formats[numpy.argmax(scores)]
    if len(lines) > 10:
        for iline, line in enumerate(lines[10:]):
            if (not (iline % 1000)) and (not progressf(float(iline)/(2*len(lines)))):
                return None
            format.add(line)
    format.finish()
    iline = 0
    tokenize = format.tokenize
    convert = format.convert
    time_col = int(bool(format.sampling))
    f_out = qubx.tree.Node('QubData')

    base, ext = os.path.splitext(path)
    sess = qubx.tree.Open(base+'.qsf', True)
    sess.close() # release file, keep in mem
    sampling = sess['Sampling'].data and sess['Sampling'].data[0] or 1e-4
    sampling = format.sampling or RequestSampling(sampling)
    f_out['Sampling'].data = sess['Sampling'].data = sampling

    f_out['SignalCount'].data = format.signal_count
    signals = f_out['Signals']
    for s in xrange(format.signal_count):
        sig = signals.append('Signal')
        sig['Name'].data = ((s + time_col) < len(format.headers)) and format.headers[s+time_col] or ('ch%d'%(s+1))
    segments = f_out['Segments']
    seg = qubx.tree.NullNode()
    prog_n = numpy.sum
    #f_out.saveAs(temppath)
    for seg_len in format.seg_lens:
        seg = segments.insert('Segment', seg)
        seg['SampleCount'].data = seg_len
        signals = seg['Signals']
        sig_storage = []
        for s in xrange(format.signal_count):
            sig = signals.append('Signal')['Samples']
            sig.data.setup(qubx.tree.QTR_TYPE_FLOAT, seg_len, 1)
            #sig.data.preload = False
            sig_storage.append(sig)
        #f_out.re_map_data() # memory maps changed location as tree was built; re-find
        for s in xrange(format.signal_count):
            sig_storage[s] = sig_storage[s].storage.data
        isample = 0
        while isample < seg_len:
            try:
                if (not (iline % 1000)) and (not progressf(0.5 + float(iline)/(len(lines)))):
                    return None
                line = lines[iline]
                iline += 1
                tokens = [x for x in [t.strip() for t in tokenize(line)] if x]
                if len(tokens) > time_col:
                    for s,x in enumerate(convert(field) for field in tokens[time_col:]):
                        sig_storage[s][isample] = x
                    isample += 1
            except ValueError:
                pass
    #f_out.clone().saveAs('foo.qtr')
    
    datafile = QubData_Scratch(path) # os.path.split(path)[1])
    datafile.tree = f_out
    #datafile.temppath = temppath

    datafile.read_session(sess, progressf)
    
    return datafile

SetReader('.txt', 'Text files', OpenTXT)
SetReader('.csv', 'Comma-separated files', OpenTXT)


def OpenDWT(path, progressf):
    data = QubData()
    data.path = path
    data.signals.append({"Name" : "Current", "Units" : "pA"})
    
    # default sampling, amps, stds
    sampling = [0.0]
    first = [0]
    segi = [0]
    class_count = 10
    amps = [1.0*i for i in xrange(class_count)]
    stds = [0.1] * class_count
    classes = []
    durations = []
    iDwell = 0
    
    def set_seg_dwells():
        if not classes:
            return
        cc = numpy.array(classes, dtype='int32')
        dd = numpy.array(durations, dtype='float32')
        ff = numpy.zeros(shape=cc.shape, dtype='int32')
        ll = numpy.zeros(shape=cc.shape, dtype='int32')
        aa = [x for x in amps]
        ss = [x for x in stds]
        sampling[0] = sampling[0] or numpy.min(dd)
        data.sampling = sampling[0] * 1e-3
        dw_samp_ct = numpy.array(dd, copy=True)
        dw_samp_ct /= sampling[0]
        dw_samp_ct += 0.5
        dw_samp_ct = dw_samp_ct.astype('int32')
        f = first[0]
        for i in xrange(len(cc)):
            nextf = f + dw_samp_ct[i]
            ff[i] = f
            ll[i] = nextf - 1
            f = nextf
        data.segmentation.add_seg(first[0], f-1, first[0]*data.sampling*1e3)
        data.ideal[0].idl.set_dwells(len(cc), ff, ll, cc)
        data.ideal[0].seg[ segi[0] ].amp = aa
        data.ideal[0].seg[ segi[0] ].std = ss
        first[0] = f
        segi[0] = segi[0] + 1
    
    hdr_pat = re.compile(r"\S+\s+(\d+)\s+\S+\s+(\d+)(.*)")
    hdr_ext_pat = re.compile(r"\s+\S+\s+([0-9.+-eE]+)\s+\S+\s+\S+\s+\S+\s+(\d+)(.*)")
    amp_std_pat = re.compile(r"\s+([0-9.+-eE]+)\s+([0-9.+-eE]+)(.*)")
    dwell_pat = re.compile(r"^\s*(\d+)\s([0-9.+-eE]+)")
    for line in open(path, "r"):
        match = dwell_pat.match(line)
        if match:
            if len(classes) <= iDwell:
                classes.append(int(match.group(1)))
                durations.append(float(match.group(2)))
            else:
                classes[iDwell] = int(match.group(1))
                durations[iDwell] = float(match.group(2))
            iDwell += 1
            continue
        match = hdr_pat.match(line)
        if match:
            set_seg_dwells()
            dwell_count = int(match.group(2))
            classes = [-1] * dwell_count
            durations = [1.0] * dwell_count
            iDwell = 0
            match = hdr_ext_pat.match(match.group(3))
            if match:
                if not sampling[0]:
                    sampling[0] = float(match.group(1))
                class_count = int(match.group(2)) 
                tail = match.group(3)
                amp = [1.0*i for i in xrange(class_count)]
                std = [0.1] * class_count
                for i in xrange(class_count):
                    match = amp_std_pat.match(tail)
                    if not match:
                        break
                    amp[i] = float(match.group(1))
                    std[i] = float(match.group(2))
                    tail = match.group(3)
    set_seg_dwells()
    
    return data

SetReader('.dwt', 'QuB idealized data', OpenDWT)
        


class QubDatas(ObjectTable):
    """A table-like collection of open files."""
    def __init__(self):
        ObjectTable.__init__(self, ' Data', global_name='QubX.Data.table')
        self.add_field('Name', "", acceptNothing, '%s', '', get=self.get_name, set=None)
        self.add_field('In Folder', '', acceptNothing, '%s', '', get=self.get_folder, set=None)
    def get_name(self, obj):
        return os.path.split(obj.path)[1]
    def get_folder(self, obj):
        return os.path.split(obj.path)[0]



if __name__ == "__main__":
    import sys
    qubx.tree.CHOOSE_FLAVOR('numpy')
    data = QubData_QDF(sys.argv[1])
    for s in xrange(data.signals.size):
        print '------------------ Signal %i --------------------------------------------' % (s+1)
        i = 0
        for f, l, n in data.segmentation.segments:
            firsts, lasts, classes, durations = data.ideal[s].idl.get_dwells(f, l, False, True)
            amp, std = data.ideal[s].seg[i].amp, data.ideal[s].seg[i].std
            print ("Segment: %i Dwells: %i Sampling(ms): %f Start(ms): %f ClassCount: %i"
                   % (i+1, len(firsts), data.sampling*1e3, f*data.sampling*1e3, len(amp))),
            for c in xrange(len(amp)):
                print amp[c], std[c],
            print
            for d in xrange(len(firsts)):
                print "\t%i\t%f" % (classes[d], durations[d])
            i += 1
        

## Read PointsPerTrace? unknown number of divs?  is idle_add late enough?

## abstract output file a la extract?
 # (channelcount, sampling, ...) scaling, dtype, time-column?, seg-columns?
 #    [onsaving: pass clone of qsf]: dc scale/off, pointsPerTrace
 # open(filename)
 # beginSegment(startTime, sampleCount)
 # writeBuffer(buf, sampleCount)
 # endSegment()
 # close
 #    [patch clone of qsf: scaling, fileext, filter*, ADDataSize, TimeColumn, Segmented, SegmentSizes, ExpCond, Idealization]
 #    [save qsf as]


# Display:PointsPerTrace
# Display:DataYM{in|ax}[i]   # Display:Channels:Channel:DataYM{in|ax} (except +/- 10,000)
# Display:FilterOn[i]
# Display:FilterFreq[i]
# FileExt = ".qdf"
# DataChannels:[Channel]:{Name|Units|Scaling|Offset}
# Sampling
# Scaling
# ADChannelCount
# ADDataSize
# TimeColumn (txt)
# Segmented (txt, True)
# SegmentSizes
# ExpCond:[*]:{<data>|ChannelIndex}
# Idealization:[Channel]:[Segment]:...


